import { FlexpdfConversion } from '@magloft/db-client'
import { BodyParams, Controller, Post, Req } from '@tsed/common'
import { BadRequest } from '@tsed/exceptions'
import { CollectionOf, Required, Returns } from '@tsed/schema'
import { FlexpdfConversionEntity, FlexpdfPreflightEntity } from '../../entities'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

export class PreflightParams {
  @Required() pdfHash: string
  @Required() @CollectionOf(String) pageHashes: string[]
}

export class ConfirmParams extends PreflightParams {
  @Required() numPages: number
}

@Controller('/conversion')
@Auth()
export class ConversionController {
  @Post('/preflight')
  @Returns(200, FlexpdfPreflightEntity)
  async preflight(@Req() { user }: AuthRequest, @BodyParams() { pdfHash, pageHashes }: PreflightParams) {
    const unconverted = await this.getUnconvertedPages(user.id, pdfHash, pageHashes)

    const response = new FlexpdfPreflightEntity()

    response.unconverted = unconverted
    response.conversionCredits = unconverted.length
    response.availableCredits = user.flexpdfCredits

    return response
  }


  @Post('/confirm')
  @Returns(200, FlexpdfConversionEntity)
  async confirm(@Req() { user }: AuthRequest, @BodyParams() { pdfHash, pageHashes, numPages }: ConfirmParams) {
    const unconverted = await this.getUnconvertedPages(user.id, pdfHash, pageHashes)
    if (unconverted.length > numPages) { unconverted.length = numPages } // only convert x amount of page
    if (user.flexpdfCredits < unconverted.length) { throw new BadRequest('Not Enough Credits') }

    for (const pageHash of unconverted) {
      const conversion = new FlexpdfConversion({ userId: user.id, pdfHash, pageHash })
      await conversion.save()
    }

    user.flexpdfCredits = user.flexpdfCredits - unconverted.length
    await user.save()

    const response = new FlexpdfConversionEntity()
    response.converted = unconverted
    response.conversionCredits = unconverted.length
    response.availableCredits = user.flexpdfCredits
    response.totalCreditsSpent = await FlexpdfConversion.count({ where: { userId: user.id } })
    return response
  }

  private async getUnconvertedPages(userId: number, pdfHash: string, pageHashes: string[]) {
    const result = []
    for (const pageHash of pageHashes) {
      const existingConversion = await FlexpdfConversion.findOne({ where: { userId, pdfHash, pageHash } })
      if (!existingConversion) {
        result.push(pageHash)
      }
    }

    return result
  }
}
