import { FlexpdfInvoice } from '@magloft/db-client'
import { BodyParams, Controller, Get, PathParams, Post, Req } from '@tsed/common'
import { BadRequest } from '@tsed/exceptions'
import { Required, Returns } from '@tsed/schema'
import { FlexpdfInvoiceEntity } from '../../entities'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'
import { InvoiceData, StripeService } from '../../services/StripeService'

const CREDITS_PER_CENT = 0.01

export class CreateInvoiceParams {
  @Required() amount: number
  @Required() credits: number
}

@Controller('/invoices')
@Auth()
export class InvoiceController {
  constructor(private stripe: StripeService) { }

  @Get()
  @Returns(200, Array).Of(FlexpdfInvoiceEntity)
  async getAll(@Req() { user }: AuthRequest): Promise<FlexpdfInvoiceEntity[]> {
    const invoices = await user.getFlexpdfInvoices()
    return FlexpdfInvoiceEntity.map(invoices)
  }

  @Get('/:id')
  @Returns(200, FlexpdfInvoiceEntity)
  async get(@PathParams('id') id: number, @Req() { user }: AuthRequest): Promise<FlexpdfInvoiceEntity> {
    const invoice = await FlexpdfInvoice.findOne({ where: { userId: user.id, id } })
    return FlexpdfInvoiceEntity.from(invoice)
  }

  @Post()
  @Returns(200, FlexpdfInvoiceEntity)
  async post(@Req() { user }: AuthRequest, @BodyParams() { amount, credits }: CreateInvoiceParams): Promise<FlexpdfInvoiceEntity> {
    // Validate Dollar VS Credits Amount
    if (credits !== amount * CREDITS_PER_CENT) { throw new BadRequest('Invalid Dollar / Credit amount') }

    // Get OR Create Stripe Customer
    user.stripeCustomer = user.stripeId
      ? await this.stripe.getCustomer(user.stripeId!)
      : await this.stripe.createCustomer({ email: user.email!, description: `Customer #${user.id}`, metadata: { 'user_id': user.id } })
    user.stripeId = user.stripeCustomer.id
    await user.save()

    // Create invoice data
    const invoiceData: InvoiceData = {
      customer: user.stripeCustomer,
      amount,
      currency: 'usd',
      description: `${credits} FlexPDF Credits`,
      collectionMethod: 'charge_automatically'
    }

    // If Payment method exists, auto-charge
    if (!user.stripeCustomer.default_source) {
      invoiceData.collectionMethod = 'send_invoice'
      invoiceData.daysUntilDue = 3
    }

    // In either case, send Invoice via Email, and get Invoice URL
    const stripeInvoice = await this.stripe.createInvoice(invoiceData)

    // Create Transaction
    const flexpdfInvoice = await user.createFlexpdfInvoice({
      source: 'stripe',
      sourceInvoiceId: stripeInvoice.id,
      amount,
      credits,
      sourceInvoice: stripeInvoice
    })

    // Pay Invoice
    await this.stripe.payInvoice(stripeInvoice)

    // Respond
    return FlexpdfInvoiceEntity.from(flexpdfInvoice)
  }
}

