import { Article, col, fn, LogicType, Op, Publication, where, WhereOptions } from '@magloft/db-client'
import { BodyParams, Controller, Get, PathParams, Post, QueryParams, Req } from '@tsed/common'
import { NotFound } from '@tsed/exceptions'
import { CollectionOf, Returns } from '@tsed/schema'
import { ArticleEntity } from '../../entities'
import { Paginate, PaginateRequest } from '../../lib/decorators/Paginate'
import { ListEntity } from '../../lib/ListEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

class ArticleListEntity extends ListEntity<ArticleEntity> {
  @CollectionOf(ArticleEntity) records
}

function buildWhere(issueId?: number, searchText?: string, categoryId?: number) {
  const options: WhereOptions[] = []
  options.push({ date: { [Op.ne]: null } })
  options.push({ html: { [Op.ne]: null } })
  if (issueId) { options.push({ issueId }) }
  if (searchText) { options.push({ title: { [Op.like]: `%${searchText}%` } }) }
  if (categoryId) { options.push(where(fn('JSON_SEARCH', col('Article.categoryIds'), 'one', categoryId), { [Op.not]: null } as unknown as LogicType)) }
  return { [Op.and]: options }
}

@Controller('/articles')
export class ArticleController {

  @Get('/:publicationId')
  @Auth()
  @Returns(200, ArticleListEntity)
  @Paginate(ArticleEntity, { orderBy: 'date', orderDir: 'desc', offset: 0, limit: 100 })
  async getAll(
    @Req() { user, offset, limit, orderBy, orderDir, paginate }: AuthRequest & PaginateRequest,
    @PathParams('publicationId') publicationId: number,
    @QueryParams('issueId') issueId?: number,
    @QueryParams('searchText') searchText?: string,
    @QueryParams('categoryId') categoryId?: number
  ): Promise<ArticleListEntity> {
    const publication = await Publication.findOne({ where: { id: publicationId, userId: user.id } })
    if (!publication) { throw new NotFound('Publication not found') }
    const query = buildWhere(issueId, searchText, categoryId)
    return paginate(
      publication.countArticles({ where: query }),
      publication.getArticles({ where: query, order: [[orderBy, orderDir]], offset, limit })
    )
  }

  @Get('/:publicationId/:id')
  @Returns(200, ArticleEntity)
  async get(@PathParams('id') id: number): Promise<ArticleEntity> {
    const user = await Article.findById(id)
    if (user == null) { throw new NotFound(`User ${id} not found`) }
    return ArticleEntity.from(user)
  }

  @Post('/:publicationId/import')
  @Auth()
  @Returns(200, Array).Of(ArticleEntity)
  async post(
    @Req() { user }: AuthRequest,
    @PathParams('publicationId') publicationId: number,
    @BodyParams('records', ArticleEntity) records: ArticleEntity[]
  ): Promise<ArticleEntity[]> {
    const publication = await Publication.findOne({ where: { userId: user.id, id: publicationId } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }
    const articles = await Promise.all(records.map(({ title, html, issueId, unlockType, position, source, cover }) => {
      return publication.createArticle({ title, html, issueId, unlockType, position, source, cover })
    }))
    return ArticleEntity.map(articles)
  }
}
