import { CustomFont, Op } from '@magloft/db-client'
import { BodyParams, Controller, Get, Post, Req } from '@tsed/common'
import { UnprocessableEntity } from '@tsed/exceptions'
import { CollectionOf, Property, Required, Returns, Status } from '@tsed/schema'
import { paramCase } from 'change-case'
import fetch from 'node-fetch'
import { CustomFontEntity } from '../../entities/CustomFontEntity'
import { parseFonts } from '../../lib/fonts'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'
import { PubSubService } from '../../services/PubSubService'

export class CreateFontParams {
  @Required() source: 'google' | 'adobe' | 'file'
  @Property() url: string
}

export class InstallFontParams {
  @Required() @CollectionOf(String) fonts: string[]
}

@Controller('/custom-fonts')
export class CustomFontController {
  constructor(private pubSub: PubSubService) { }

  @Get()
  @Auth('anonymous')
  @Returns(200, Array).Of(CustomFontEntity)
  async getAll(@Req() { user }: AuthRequest): Promise<CustomFontEntity[]> {
    const customFonts = await CustomFont.findAll({
      where: user
        ? { [Op.or]: [{ public: true }, { identifier: user.properties.get('customFonts') ?? [] }] }
        : { public: true },
      order: [['identifier', 'asc']]
    })
    return CustomFontEntity.map(customFonts)
  }

  @Post('/')
  @Auth('user')
  @Returns(200, Array).Of(CustomFontEntity)
  async post(@Req() { user }: AuthRequest, @BodyParams(CreateFontParams) params: CreateFontParams): Promise<CustomFontEntity[]> {
    let entities: CustomFontEntity[]

    // Import Custom Font
    if (params.source === 'google') {
      entities = await this.importGoogleFont(params)
    } else if (params.source === 'adobe') {
      entities = await this.importAdobeFont(params)
    } else if (params.source === 'file') {
      entities = await this.importFileFont(params)
    } else {
      throw new UnprocessableEntity(`Invalid font source: '${params.source}'`)
    }

    // Assign Custom Font
    for (const entity of entities) {
      const customFonts: string[] = user.properties.get('customFonts') ?? []
      if (!customFonts.includes(entity.identifier)) { user.properties.set('customFonts', [...customFonts, entity.identifier]) }
    }
    await user.save()

    // Return Custom Font Entity
    return entities
  }

  @Post('/install')
  @Auth()
  @Status(204)
  async postInstall(@Req() { user }: AuthRequest, @BodyParams() { fonts }: InstallFontParams) {
    const installedFonts = user.properties.get('customFonts') as string[] || []
    const customFonts = await CustomFont.findAll({ where: { identifier: fonts, public: false } })
    customFonts.forEach(font => {
      if (!installedFonts.includes(font.identifier)) {
        installedFonts.push(font.identifier)
      }
    })
    user.properties.set('customFonts', installedFonts)
    await user.save()
    return true
  }

  private async importGoogleFont({ url }: CreateFontParams): Promise<CustomFontEntity[]> {
    const match = url.match(/^https:\/\/fonts\.google\.com\/specimen\/([-a-zA-Z0-9_\+~#=]{2,256})$/)
    if (!match) { throw new UnprocessableEntity('Invalid font url') }
    const title = decodeURI(match[1]).replace(/\+/g, ' ')
    const identifier = paramCase(title)
    const [customFont, created] = await CustomFont.findOrCreate({ where: { identifier }, defaults: { public: false, variants: [] } })
    if (created) { await this.pubSub.trigger({ name: 'import-font', data: { url } }) }
    return [CustomFontEntity.from(customFont)]
  }

  private async importAdobeFont({ url }: CreateFontParams): Promise<CustomFontEntity[]> {
    // Validate URL
    const match = url.match(/^https:\/\/use\.typekit\.net\/[-a-zA-Z0-9_\+~#=]+\.css$/)
    if (!match) { throw new UnprocessableEntity('Invalid font url') }

    // Retrieve Web Project Fonts
    const respone = await fetch(url)
    const text = await respone.text()
    const fonts = parseFonts(text)
    if (fonts.length > 10) { throw new UnprocessableEntity(`A maximum of 10 fonts can be imported at a time. Your project currencly contains ${fonts.length} fonts.`) }

    // Create and Retrieve Custm Fonts
    let shouldImport = false
    const customFonts = await Promise.all(fonts.map(async (font) => {
      const [customFont, initialized] = await CustomFont.findOrBuild({ where: { identifier: font.identifier }, defaults: { variants: [], public: false } })
      if (initialized) { shouldImport = true }
      return customFont.save()
    }))

    // Trigger Font Import Job
    if (shouldImport) { await this.pubSub.trigger({ name: 'import-font', data: { url } }) }

    // Return Custom Fonts
    return CustomFontEntity.map(customFonts)
  }

  private async importFileFont({ url }: CreateFontParams): Promise<CustomFontEntity[]> {
    const match = url.match(/^https:\/\/cdn\.magloft\.(com|dev)\/secure\/[0-9]+\/([-a-zA-Z0-9_\+~#=]{2,256})\//)
    if (!match) { throw new UnprocessableEntity('Invalid font url') }
    const identifier = match[2]
    const [customFont, created] = await CustomFont.findOrCreate({ where: { identifier }, defaults: { public: false, variants: [] } })
    if (created) { await this.pubSub.trigger({ name: 'import-font', data: { url } }) }
    return [CustomFontEntity.from(customFont)]
  }
}
