import { Issue, Publication } from '@magloft/db-client'
import { BodyParams, Controller, Get, PathParams, Post, QueryParams, Req } from '@tsed/common'
import { NotFound } from '@tsed/exceptions'
import { CollectionOf, Default, Enum, Property, Required, Returns } from '@tsed/schema'
import { IssueEntity, IssueGenerator, IssueSource, IssueUnlockType } from '../../entities'
import { Paginate, PaginateRequest } from '../../lib/decorators/Paginate'
import { ListEntity } from '../../lib/ListEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

class IssueListEntity extends ListEntity<IssueEntity> {
  @CollectionOf(IssueEntity) records
}

class CreateIssueParams {
  @Required() title: string
  @Property() info: string
  @Required() @Enum(IssueSource) source: IssueSource
  @Property() @Default(IssueUnlockType.FREE) @Enum(IssueUnlockType) unlockType: IssueUnlockType
  @Property() @Default('defaults/images/cover.jpg') cover: string
  @Property() @Enum(IssueGenerator) generator?: IssueGenerator
}

@Controller('/issues')
export class IssueController {

  @Get('/:publicationId')
  @Auth()
  @Returns(200, IssueListEntity)
  @Paginate(IssueEntity, { orderBy: 'date', orderDir: 'desc', offset: 0, limit: 100 })
  async getAll(
    @Req() { user, offset, limit, orderBy, orderDir, paginate }: AuthRequest & PaginateRequest,
    @PathParams('publicationId') publicationId: number,
    @QueryParams('generator') generator?: string
  ): Promise<IssueListEntity> {
    const publication = await Publication.findOne({ where: { userId: user.id, id: publicationId } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }
    const where = generator ? { generator } : {}
    return paginate(publication.countIssues({ where }), publication.getIssues({ where, order: [[orderBy, orderDir]], offset, limit }))
  }

  @Get('/:publicationId/:id')
  @Auth()
  @Returns(200, IssueEntity)
  async get(@Req() { user }: AuthRequest, @PathParams('publicationId') publicationId: number, @PathParams('id') id: number): Promise<IssueEntity> {
    const publication = await Publication.findOne({ where: { userId: user.id, id: publicationId } })
    if (!publication) { throw new NotFound(`Publication ${publicationId} not found`) }
    const issue = await Issue.findOne({ where: { id, publicationId } })
    if (!issue) { throw new NotFound(`Issue ${id} not found`) }
    return IssueEntity.from(issue)
  }

  @Post('/:publicationId')
  @Auth()
  @Returns(200, IssueEntity)
  async post(@Req() { user }: AuthRequest, @BodyParams(CreateIssueParams) params: CreateIssueParams, @PathParams('publicationId') publicationId: number): Promise<IssueEntity> {
    const publication = await Publication.findOne({ where: { userId: user.id, id: publicationId } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }
    const name = `issue_${publication.userId}_${Math.round(new Date().getTime() / 1000)}`
    const issue = await publication.createIssue({ userId: user.id, ...params, name })
    return IssueEntity.from(issue)
  }
}
