import { Publication } from '@magloft/db-client'
import { AccessLevel } from '@magloft/db-client/lib/models/User'
import { Controller, Get, PathParams, QueryParams, Req } from '@tsed/common'
import { BadRequest, NotFound } from '@tsed/exceptions'
import { CollectionOf, Returns } from '@tsed/schema'
import { PublicationEntity } from '../../entities'
import { ListEntity } from '../../lib/ListEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

class PublicationListEntity extends ListEntity<PublicationEntity> {
  @CollectionOf(PublicationEntity) records
}

@Controller('/publications')
export class PublicationController {
  @Get()
  @Auth()
  @Returns(200, PublicationListEntity)
  async getAll(
    @Req() { user }: AuthRequest,
    @QueryParams('offset', Number) offset = 0,
    @QueryParams('limit', Number) limit = 100,
    @QueryParams('orderBy') orderBy: 'id' | 'title' | 'createdAt' = 'createdAt',
    @QueryParams('orderDir') orderDir: 'asc' | 'desc' = 'desc'
  ): Promise<PublicationListEntity> {
    const [total, publications] = await Promise.all([
      user.countPublications(),
      user.getPublications({ order: [[orderBy, orderDir]], offset, limit })
    ])
    return { total, offset, limit, records: PublicationEntity.map(publications) }
  }

  @Get('/managed')
  @Auth()
  @Returns(200, Array).Of(PublicationEntity)
  async getManaged(@Req() { user }: AuthRequest, @QueryParams('role') role: AccessLevel = 'author'): Promise<PublicationEntity[]> {
    if (!['author', 'editor', 'publisher'].includes(role)) { throw new BadRequest(`Invalid role '${role}'`) }
    const publications = await user.getManagedPublications(role)
    return PublicationEntity.map(publications)
  }

  @Get('/:id')
  @Auth()
  @Returns(200, PublicationEntity)
  async get(@Req() { user }: AuthRequest, @PathParams('id') id: number): Promise<PublicationEntity> {
    const publication = await Publication.findOne({ where: { userId: user.id, id } })
    if (publication == null) { throw new NotFound(`Publication ${id} not found`) }
    return PublicationEntity.from(publication)
  }
}
