import { Op, Publication, Reader } from '@magloft/db-client'
import { Controller, Get, PathParams, Req } from '@tsed/common'
import { NotFound } from '@tsed/exceptions'
import { CollectionOf, Returns } from '@tsed/schema'
import { ReaderEntity } from '../../entities/ReaderEntity'
import { Paginate, PaginateRequest } from '../../lib/decorators/Paginate'
import { ListEntity } from '../../lib/ListEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

class ReaderListEntity extends ListEntity<ReaderEntity> {
  @CollectionOf(ReaderEntity) records
}

@Controller('/readers')
export class ReaderController {
  constructor() { }

  @Get('/:publicationId')
  @Auth()
  @Returns(200, ReaderListEntity)
  @Paginate(ReaderEntity, { orderBy: 'createdAt', orderDir: 'desc', offset: 0, limit: 100 })
  async getAll(@Req() { offset, limit, orderBy, orderDir, paginate }: AuthRequest & PaginateRequest, @PathParams('publicationId') publicationId: number
  ): Promise<ReaderListEntity> {
    // Retrieve Publication
    const publication = await Publication.findOne({ where: { id: publicationId } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }

    // Return Reader Pagination
    return paginate(publication.countReaders(), publication.getReaders({ order: [[orderBy, orderDir]], offset, limit }))
  }

  @Get('/:publicationId/:emailOrId')
  @Auth()
  @Returns(200, ReaderEntity)
  async getByEmailOrId(@PathParams('publicationId') publicationId: number, @PathParams('emailOrId') emailOrId: string): Promise<ReaderEntity> {
    // Retrieve Publication
    const reader = await Reader.findOne({ where: { publicationId, [Op.or]: [{ email: emailOrId }, { externalId: emailOrId }, { id: emailOrId }] } })
    if (reader == null) { throw new NotFound(`Reader for ${emailOrId} not found`) }

    // Return Reader Pagination
    return ReaderEntity.from(reader)
  }
}
