import { Reader, ReaderSubscription } from '@magloft/db-client'
import { EmailTemplate } from '@magloft/db-client/lib/models/EmailTemplate'
import { BodyParams, Controller, PathParams, Post, Req } from '@tsed/common'
import { Conflict } from '@tsed/exceptions'
import { Any, CollectionOf, Description, Email, MaxLength, MinLength, Property, Required, Returns, Schema } from '@tsed/schema'
import { ReaderEntity } from '../../entities/ReaderEntity'
import { ReaderSubscriptionEntity } from '../../entities/ReaderSubscriptionEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'
import { AuthService } from '../../services/AuthService'
import { MailService } from '../../services/MailService'

@Schema({ title: 'Reader With Subscription' })
export class ReaderWithSubscriptionEntity {
  @Property() @Description('Status') status: string
  @Property(ReaderEntity) @Description('Reader') reader: ReaderEntity
  @Property(ReaderSubscriptionEntity) @Description('Subscription') subscription: ReaderSubscriptionEntity
}

export class CreateReaderSubscriptionParams {
  @Required() @Email() email: string
  @Property() sendEmail: boolean = false
  @Property() confirm: boolean = true
  @Property() @MinLength(3) @MaxLength(32) password?: string
  @Property() externalId?: string
  @Property() firstname?: string
  @Property() lastname?: string
  @Property() orderId?: string
  @Any(String, Number) startDate?: Date
  @Any(String, Number) endDate?: Date
  @Property() active?: boolean
  @Property() customData?: { [key: string]: string }
  @CollectionOf(String) classificationIds?: string[]
}

@Controller('/reader-subscriptions')
export class ReaderSubscriptionController {
  constructor(private auth: AuthService, private mail: MailService) { }

  @Post('/:publicationId')
  @Auth('user')
  @Returns(200, ReaderWithSubscriptionEntity)
  async post(@Req() { user }: AuthRequest, @PathParams('publicationId') publicationId: number, @BodyParams(CreateReaderSubscriptionParams) params: CreateReaderSubscriptionParams): Promise<ReaderWithSubscriptionEntity> {
    const publication = await user.getManagedPublication(publicationId, 'publisher')
    if (!publication) { throw new Conflict(`Could not find publication with ID ${publicationId}`) }

    // Encrypt Password
    const createdAt = new Date()
    const password = params.password ?? this.auth.getDefaultPassword(params.email, createdAt)
    const encryptedPassword = this.auth.encryptPassword(password)
    const authenticationToken = this.auth.generateFriendlyToken(16)

    // Find or create reader
    const defaults = { createdAt, encryptedPassword, authenticationToken, classificationIds: [] }
    const [reader, readerInitialized] = params.externalId
      ? await Reader.findOrBuild({ where: { externalId: params.externalId, publicationId }, defaults })
      : await Reader.findOrBuild({ where: { email: params.email, publicationId }, defaults })
    reader.email = params.email
    if ('customData' in params) { reader.customData = Object.assign({}, reader.customData, params.customData) }
    if ('classificationIds' in params) {
      reader.classificationIds = params.classificationIds.reduce((arr: string[], value: string) => {
        if (!arr.includes(value)) { arr.push(value) }
        return arr
      }, reader.classificationIds ?? [])
    }
    if ('firstname' in params) { reader.name = params.firstname }
    if ('lastname' in params) { reader.lastname = params.lastname }
    if (params.confirm) { reader.confirmedAt = createdAt }
    await reader.save()

    // Create subscription
    const [subscription, initialized] = await ReaderSubscription.findOrBuild({ where: { orderId: params.orderId || null, source: 'custom', publicationId, readerId: reader.id, active: true } })
    if ('startDate' in params) { subscription.startDate = params.startDate }
    if ('endDate' in params) { subscription.endDate = params.endDate }
    if ('active' in params) { subscription.active = params.active }
    await subscription.save()

    // Manage Confirmation / Email Delivery
    if (readerInitialized && params.sendEmail) {
      const emailTemplate = await EmailTemplate.findOne({ where: { publicationId, active: true, action: 'confirmation_instructions' } })
      if (emailTemplate) { await this.mail.sendReaderEmail(publication, reader, emailTemplate, { 'reader.password': password }) }
    }

    // Create and return entity
    const entity = new ReaderWithSubscriptionEntity()
    entity.status = initialized ? 'created' : 'updated'
    entity.reader = ReaderEntity.from(reader, { password })
    entity.subscription = ReaderSubscriptionEntity.from(subscription)
    return entity
  }
}
