import { User } from '@magloft/db-client'
import * as mms from '@magloft/mms-db'
import { BodyParams, Configuration, Controller, Delete, Get, PathParams, Post, Put, Req, Res } from '@tsed/common'
import { Conflict, NotFound } from '@tsed/exceptions'
import { Email, MaxLength, MinLength, Property, Required, Returns } from '@tsed/schema'
import { createHmac } from 'crypto'
import { SessionEntity, UserEntity } from '../../entities'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'
import { AuthService } from '../../services/AuthService'

export class CreateUserParams {
  @Required() @Email() email: string
  @Required() @MinLength(3) @MaxLength(32) password: string
  @Property() firstname?: string
  @Property() lastname?: string
}

export class UpdateUserParams {
  @Property() countryCode?: string
  @Property() vatNumber?: string
}

export class CreateUserSessionParams {
  @Required() @Email() email: string
  @Required() @MinLength(3) @MaxLength(32) password: string
}

export interface IntercomConfiguration {
  secretKey: string
}

@Controller('/users')
export class UserController {
  private intercomSecret: string

  constructor(private auth: AuthService, @Configuration() private configuration: Configuration) {
    this.intercomSecret = this.configuration.get<IntercomConfiguration>('intercom').secretKey
  }

  @Get('/session')
  @Auth('user')
  @Returns(200, SessionEntity)
  async getSession(@Res() response: Res, @Req() { user }: AuthRequest): Promise<SessionEntity> {
    this.auth.setCookie(response, user)
    const session = new SessionEntity()
    const userEntity = UserEntity.from(user)
    session.user = userEntity
    session.intercomUser = this.userIntercomDetails(userEntity)
    session.token = this.auth.encrypt(user)
    session.authenticated = true
    return session
  }

  @Post('/session')
  @Returns(200, SessionEntity)
  async postSession(@Res() response: Res, @BodyParams() { email, password }: CreateUserSessionParams): Promise<SessionEntity> {
    const session = new SessionEntity()
    session.authenticated = false
    const user = await User.findOne({ where: { email } })
    if (!user || !user.email) { return session }
    if (!this.auth.validatePassword(user.encryptedPassword, password)) { return session }
    this.auth.setCookie(response, user)
    const userEntity = UserEntity.from(user)
    session.user = userEntity
    session.intercomUser = this.userIntercomDetails(userEntity)
    session.token = this.auth.encrypt(user)
    session.authenticated = true
    return session
  }

  @Delete('/session')
  @Auth()
  @Returns(200, SessionEntity)
  async deleteSession(@Res() response: Res): Promise<SessionEntity> {
    this.auth.clearCookie(response)
    const session = new SessionEntity()
    session.authenticated = false
    return session
  }

  @Post()
  @Returns(200, SessionEntity)
  async post(@Res() response: Res, @BodyParams(CreateUserParams) { email, firstname, lastname, password }: CreateUserParams): Promise<SessionEntity> {
    const session = new SessionEntity()
    session.authenticated = false
    const existingUser = await User.findOne({ where: { email } })
    if (existingUser) { throw new Conflict('Email already registered') }
    const encryptedPassword = this.auth.encryptPassword(password)
    const mmsUser = (await mms.User.findOne({ where: { email } })) || (await mms.User.create({ email, admin: false }))
    const user = await User.create({ email, firstname, lastname, encryptedPassword, flexpdfCredits: 100, properties: { 'mms_id': mmsUser.id, 'mms_token': mmsUser.token } })
    await user.reload()
    const publication = await user.createPublication({
      title: 'My Publication',
      developmentMode: true,
      subscriptionStatus: 'inactive',
      publishingStatus: 'pending',
      banners: [{ size: 'cover', image: null, 'background_color': '#216CA0' }],
      storageProvider: 'magloft'
    })
    await publication.update({
      appId: `com.magloft.magazine.${publication.id}`,
      slug: `magazine-${publication.id}`
    })
    this.auth.setCookie(response, user)
    const userEntity = UserEntity.from(user)
    session.user = userEntity
    session.intercomUser = this.userIntercomDetails(userEntity)
    session.token = this.auth.encrypt(user)
    session.authenticated = true
    return session
  }

  @Put()
  @Auth()
  @Returns(200, SessionEntity)
  async put(
    @Req() { user }: AuthRequest,
    @BodyParams(UpdateUserParams) { vatNumber, countryCode }: UpdateUserParams
  ): Promise<SessionEntity> {

    if (vatNumber || countryCode) {
      if (vatNumber) { user.properties.set('vatNumber', vatNumber) }
      if (countryCode) { user.properties.set('countryCode', countryCode) }

      // TODO: Improve change detection
      user.changed('properties', true)
      await user.save()
    }


    const session = new SessionEntity()
    const userEntity = UserEntity.from(user)

    session.user = userEntity
    session.intercomUser = this.userIntercomDetails(userEntity)
    session.token = this.auth.encrypt(user)
    session.authenticated = true

    return session
  }

  @Get()
  @Auth('admin')
  @Returns(200, Array).Of(UserEntity)
  async get(): Promise<UserEntity[]> {
    const users = await User.findAll({ limit: 10 })
    return UserEntity.map(users)
  }

  @Get('/:id')
  @Auth('admin')
  @Returns(200, UserEntity)
  async getId(@PathParams('id') id: number): Promise<UserEntity> {
    const user = await User.findById(id)
    if (user == null) { throw new NotFound(`User ${id} not found`) }
    return UserEntity.from(user)
  }

  userIntercomDetails(user: UserEntity) {
    return {
      'user_id': user.id.toString(),
      'user_hash': createHmac('sha256', this.intercomSecret).update(user.id.toString()).digest('hex'),
      'name': user.fullname,
      'email': user.email
    }
  }
}
