import { Publication, PushDataChrome, PushDataSafari, PushProvider, Pwa, PwaPushSubscription } from '@magloft/db-client'
import { $log, BodyParams, Configuration, Controller, Delete, Get, PathParams, Post, Req, Res } from '@tsed/common'
import { NotFound, UnprocessableEntity } from '@tsed/exceptions'
import { CollectionOf, ContentType, Produces, Property, Required, Returns } from '@tsed/schema'
import { readFileSync } from 'fs'
import { join, resolve } from 'path'
import * as pushpackage from 'pushpackage'
import { PushDeliveryEntity } from '../../entities/PushDeliveryEntity'
import { PushSubscriptionEntity } from '../../entities/PushSubscriptionEntity'
import { PushWebsiteJsonEntity } from '../../entities/PushWebsiteJsonEntity'
import { Paginate, PaginateRequest } from '../../lib/decorators/Paginate'
import { ListEntity } from '../../lib/ListEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'
import { PushService } from '../../services/PushService'

class PushSubscriptionListEntity extends ListEntity<PushSubscriptionEntity> {
  @CollectionOf(PushSubscriptionEntity) records
}

export interface PushNotificationAction {
  action: string
  title: string
  icon?: string
  data: { [key: string]: string | number | boolean }
}

export class CreatePushSubscriptionParams {
  @Required() provider: PushProvider
  @Required() data: PushDataChrome | PushDataSafari
  @Property() readerId?: number
}

export class CreatePushDeliveryParams {
  @Required() title: string
  @Required() body: string
  @Property() data?: { [key: string]: string | number | boolean }
  @CollectionOf(Object) actions?: PushNotificationAction[]
}

@Controller('/push')
export class PushController {
  constructor(@Configuration() private configuration: Configuration, private push: PushService) { }

  @Get('/:pwaId/subscriptions')
  @Auth()
  @Returns(200, PushSubscriptionListEntity)
  @Paginate(PushSubscriptionEntity, { orderBy: 'createdAt', orderDir: 'desc', offset: 0, limit: 100 })
  async getAll(
    @Req() { user, offset, limit, orderBy, orderDir, paginate }: AuthRequest & PaginateRequest,
    @PathParams('pwaId') pwaId: number
  ): Promise<PushSubscriptionListEntity> {
    const pwa = await Pwa.findOne({ where: { id: pwaId } })
    if (!pwa) { throw new NotFound('PWA not found') }
    const publication = await Publication.findOne({ where: { id: pwa.publicationId, userId: user.id } })
    if (!publication) { throw new NotFound('Publication not found') }
    return paginate(pwa.countPwaPushSubscriptions(), pwa.getPwaPushSubscriptions({ order: [[orderBy, orderDir]], offset, limit }))
  }

  @Post('/:pwaId/subscriptions')
  @Returns(200, PushSubscriptionEntity)
  async postSubscription(@BodyParams(CreatePushSubscriptionParams) { provider, data, readerId }: CreatePushSubscriptionParams, @PathParams('pwaId') pwaId: number): Promise<PushSubscriptionEntity> {
    const pwa = await Pwa.findOne({ where: { id: pwaId } })
    if (!pwa) { throw new NotFound('PWA not found') }
    const token = this.push.getToken(provider, data)
    const [subscription] = await PwaPushSubscription.findOrCreate({ where: { pwaId, token }, defaults: { provider, data, readerId } })
    if (!subscription.readerId && readerId) {
      subscription.readerId = readerId
      await subscription.save()
    }
    return PushSubscriptionEntity.from(subscription)
  }

  @Post('/:pwaId/delivery')
  @Auth()
  @Returns(200, PushDeliveryEntity)
  async postDelivery(@Req() { user }: AuthRequest, @BodyParams(CreatePushDeliveryParams) params: CreatePushDeliveryParams, @PathParams('pwaId') pwaId: number): Promise<PushDeliveryEntity> {
    const pwa = await Pwa.findOne({ where: { id: pwaId } })
    if (!pwa) { throw new NotFound('PWA not found') }
    const publication = await Publication.findOne({ where: { id: pwa.publicationId, userId: user.id } })
    if (!publication) { throw new NotFound('Publication not found') }
    const subscriptions = await pwa.getPwaPushSubscriptions()
    const results = await Promise.all(subscriptions.map((subscription) => {
      return this.push.sendNotification(params, subscription, pwa)
    }))
    const entity = new PushDeliveryEntity()
    entity.success = true
    entity.successCount = results.filter((success) => success === true).length
    entity.errorCount = results.filter((success) => success === false).length
    return entity
  }

  @Post('/v2/pushPackages/:websitePushId')
  @ContentType('application/json')
  @Produces('application/zip')
  async postPushPackage(@PathParams('websitePushId') websitePushId: string, @Req() req: Req, @Res() res: Res): Promise<void> {
    // Retrieve PWA
    const domain = this.push.getDomainFromWebsitePushId(websitePushId)
    const pwa = await Pwa.findOne({ where: { domain } })
    if (!pwa) { throw new NotFound('PWA not found') }

    // Create Zip Entries
    const entries: pushpackage.PushPackageEntry[] = []
    for (const baseName of ['icon_16x16.png', 'icon_16x16@2x.png', 'icon_32x32.png', 'icon_32x32@2x.png', 'icon_128x128.png', 'icon_128x128@2x.png']) {
      const assetPath = resolve(join(__dirname, '..', '..', 'assets', 'pushpackage', 'icon.iconset', baseName))
      entries.push({ name: `icon.iconset/${baseName}`, buffer: readFileSync(assetPath) })
    }

    // Create website.json
    const readerId = ('readerId' in req.body) ? parseInt(req.body.readerId, 10) : 0
    const authenticationToken = this.push.encodeAuthenticationToken(pwa.id, readerId)
    const websiteJson = PushWebsiteJsonEntity.fromPwa(pwa, this.configuration.env, authenticationToken)
    entries.push({ name: 'website.json', buffer: websiteJson.toBuffer() })

    // Create Push Package
    res.attachment(`${websitePushId}.pushpackage`)
    const { key, certificate } = pwa.config.push.apple as { key: string, certificate: string }
    await pushpackage.pipe(entries, res, { key, crt: certificate })
  }

  @Post('/v1/devices/:deviceToken/registrations/:websitePushId')
  async postRegistration(@PathParams('deviceToken') deviceToken: string, @PathParams('websitePushId') websitePushId: string, @Req() req: Req): Promise<void> {
    // Retrieve PWA
    const domain = this.push.getDomainFromWebsitePushId(websitePushId)
    const pwa = await Pwa.findOne({ where: { domain } })
    if (!pwa) { throw new NotFound('PWA not found') }

    // Get Reader Id
    const authenticationToken = req.header('authorization').split(' ')[1]
    const { pwaId, readerId } = this.push.decodeAuthenticationToken(authenticationToken)
    if (pwa.id !== pwaId) { throw new UnprocessableEntity('Invalid Authentication Token payload') }

    // Store Token
    const provider = 'safari'
    const data: PushDataSafari = { deviceToken }
    const token = this.push.getToken('safari', data)
    const [subscription, created] = await PwaPushSubscription.findOrCreate({ where: { pwaId, token }, defaults: { provider, data, readerId } })
    if (!created && !subscription.readerId && readerId) {
      subscription.readerId = readerId
      await subscription.save()
    }
  }

  @Delete('/v1/devices/:deviceToken/registrations/:websitePushId')
  async deleteRegistration(@PathParams('deviceToken') deviceToken: string, @PathParams('websitePushId') websitePushId: string, @Req() req: Req): Promise<void> {
    console.log(deviceToken, websitePushId, req)
    // Retrieve PWA
    const domain = this.push.getDomainFromWebsitePushId(websitePushId)
    const pwa = await Pwa.findOne({ where: { domain } })
    if (!pwa) { return }

    // Retrieve PWA Push Subscription
    const token = this.push.getToken('safari', { deviceToken })
    const subscription = await PwaPushSubscription.findOne({ where: { pwaId: pwa.id, token } })
    if (!subscription) { return }

    // Delete PWA Push Subscription
    await subscription.destroy()
  }

  @Post('/v1/log')
  @ContentType('application/json')
  async postLogV1(@Req() req: Req): Promise<void> {
    for (const message of req.body.logs) {
      $log.info('[PushController] Safari Log', message)
    }
  }
}
