import { Article, Issue, Op, Pwa } from '@magloft/db-client'
import { Configuration, Controller, Get, PathParams, Res } from '@tsed/common'
import { NotFound } from '@tsed/exceptions'
import { SitemapStream, streamToPromise } from 'sitemap'
import { createGzip } from 'zlib'

export type IssueWithArticles = Issue & { Articles: Article[] }

@Controller('/sitemap')
export class SitemapController {
  constructor(@Configuration() private configuration: Configuration) { }

  @Get('/:domain.xml')
  async getSitemap(@PathParams('domain') domain: string, @Res() res: Res): Promise<void> {
    res.header('Content-Type', 'application/xml')
    res.header('Content-Encoding', 'gzip')

    // Fetch PWA
    const pwa = await Pwa.findOne({ where: { domain } })
    if (!pwa) { throw new NotFound('PWA not found') }

    // Fetch Articles
    const articles = await Article.findAll({
      attributes: ['id', 'slug', 'updatedAt'],
      where: { publicationId: pwa.publicationId, standalone: true, html: { [Op.ne]: null }, date: { [Op.ne]: null } },
      order: [['date', 'desc']]
    })

    // Fetch Issues
    const issues = await Issue.findAll({
      attributes: ['id', 'updatedAt'],
      where: { publicationId: pwa.publicationId, date: { [Op.ne]: null } },
      order: [['date', 'desc']],
      include: [{ model: Article, as: 'Articles', required: false, attributes: ['id', 'updatedAt'] }]
    }) as IssueWithArticles[]

    // Generate Sitemap
    const sitemap = new SitemapStream({ hostname: `https://${pwa.domain}/` })
    const pipeline = sitemap.pipe(createGzip())
    sitemap.write({ url: '/', changefreq: 'daily', priority: 0.5 })
    sitemap.write({ url: '/articles', changefreq: 'daily', priority: 0.5 })
    for (const article of articles) {
      sitemap.write({ url: `/articles/${article.slug ?? `article-${article.id}`}`, changefreq: 'weekly', priority: 0.6, lastmod: article.updatedAt })
    }
    sitemap.write({ url: '/issues', changefreq: 'daily', priority: 0.5 })
    for (const issue of issues) {
      sitemap.write({ url: `/issues/${issue.slug}`, changefreq: 'weekly', priority: 0.6, lastmod: issue.updatedAt })
      for (const article of issue.Articles) {
        sitemap.write({ url: `/issues/${issue.slug}/${article.slug ?? `article-${article.id}`}`, changefreq: 'weekly', priority: 0.6, lastmod: article.updatedAt })
      }
    }
    const promise = streamToPromise(pipeline)
    sitemap.end()
    pipeline.pipe(res).on('error', (error) => { throw error })
    await promise
  }
}
