import { Publication, ReaderSubscription } from '@magloft/db-client'
import { $log, BodyParams, Controller, Post } from '@tsed/common'
import { Name, Property, Required } from '@tsed/schema'
import { generateLegacyReceiptData } from '../../lib/iap'

type NotificationType =
  | 'CANCEL'
  | 'DID_CHANGE_RENEWAL_PREF'
  | 'DID_CHANGE_RENEWAL_STATUS'
  | 'DID_FAIL_TO_RENEW'
  | 'DID_RECOVER'
  | 'INITIAL_BUY'
  | 'INTERACTIVE_RENEWAL'
  | 'RENEWAL'

export class PendingRenewalInfo {
  @Property() @Name('auto_renew_product_id') autoRenewProductId: string
  @Property() @Name('auto_renew_status') autoRenewStatus: number
  @Property() @Name('expiration_intent') expirationIntent: number
  @Property() @Name('grace_period_expires_date') gracePeriodExpiresDate: string
  @Property() @Name('grace_period_expires_date_ms') gracePeriodExpiresDateMs: string
  @Property() @Name('grace_period_expires_date_pst') gracePeriodExpiresDatePst: string
  @Property() @Name('is_in_billing_retry_period') isInBillingRetryPeriod: number
  @Property() @Name('original_transaction_id') originalTransactionId: string
  @Property() @Name('price_consent_status') priceConsentStatus: string
  @Property() @Name('product_id') productId: string
}

export class LatestReceiptInfo {
  @Property() @Name('bid') bid: string
  @Property() @Name('bvrs') bvrs: string
  @Property() @Name('unique_identifier') uniqueIdentifier: string
  @Property() @Name('app_item_id') appItemId: string
  @Property() @Name('expires_date') expiresDate: string
  @Property() @Name('expires_date_formatted') expiresDateFormatted: string
  @Property() @Name('expires_date_formatted_pst') expiresDateFormattedPst: string
  @Property() @Name('is_in_intro_offer_period') isInIntroOfferPeriod: string
  @Property() @Name('is_trial_period') isTrialPeriod: string
  @Property() @Name('item_id') itemId: string
  @Property() @Name('original_purchase_date') originalPurchaseDate: string
  @Property() @Name('original_purchase_date_ms') originalPurchaseDateMs: string
  @Property() @Name('original_purchase_date_pst') originalPurchaseDatePst: string
  @Property() @Name('original_transaction_id') originalTransactionId: string
  @Property() @Name('product_id') productId: string
  @Property() @Name('purchase_date') purchaseDate: string
  @Property() @Name('purchase_date_ms') purchaseDateMs: string
  @Property() @Name('purchase_date_pst') purchaseDatePst: string
  @Property() @Name('quantity') quantity: string
  @Property() @Name('transaction_id') transactionId: string
  @Property() @Name('web_order_line_item_id') webOrderLineItemId: string
  @Property() @Name('unique_vendor_identifier') uniqueVendorIdentifier: string
}

export class UnifiedReceipt {
  @Property() @Name('environment') @Required() environment: 'Sandbox' | 'Production'
  @Property() @Name('latest_receipt') latestReceipt?: string
  @Property() @Name('latest_receipt_info') latestReceiptInfo?: LatestReceiptInfo[]
  @Property() @Name('pending_renewal_info') pendingRenewalInfo?: PendingRenewalInfo[]
  @Property() @Name('status') @Required() status: number
}

export class ReceiptNotification {
  @Property() @Name('notification_type') @Required() notificationType: NotificationType
  @Property() @Name('environment') @Required() environment: 'Sandbox' | 'Prod'
  @Property() @Name('auto_renew_product_id') @Required() autoRenewProductId: string
  @Property() @Name('auto_renew_status') @Required() autoRenewStatus: string
  @Property() @Name('password') @Required() password: string
  @Property() @Name('unified_receipt') @Required() unifiedReceipt: UnifiedReceipt
  @Property() @Name('latest_receipt') latestReceipt?: string
  @Property() @Name('latest_receipt_info') latestReceiptInfo?: LatestReceiptInfo
  @Property() @Name('auto_renew_status_change_date') autoRenewStatusChangeDate?: string
  @Property() @Name('auto_renew_status_change_date_ms') autoRenewStatusChangeDateMs?: string
  @Property() @Name('auto_renew_status_change_date_pst') autoRenewStatusChangeDatePst?: string
  @Property() @Name('expiration_intent') expirationIntent?: number
  @Property() @Name('latest_expired_receipt') latestExpiredReceipt?: string
  @Property() @Name('latest_expired_receipt_info') latestExpiredReceiptInfo?: LatestReceiptInfo
}

@Controller('/app-store')
export class AppStoreWebhookController {
  @Post()
  async post(@BodyParams(ReceiptNotification) { notificationType, password, latestReceipt, latestExpiredReceipt, latestReceiptInfo, latestExpiredReceiptInfo, unifiedReceipt }: ReceiptNotification): Promise<void> {
    $log.info('[AppStoreWebhookController] Received Notification', notificationType)

    // Validate Receipt
    const receipt = latestReceiptInfo ?? latestExpiredReceiptInfo
    const encodedReceipt = latestReceipt ?? latestExpiredReceipt
    if (receipt == null || encodedReceipt == null) { $log.error('Receipt not found'); return }

    // Load and validate Publication / Shared Secret
    const publication = await Publication.findOne({ where: { appId: receipt.bid } })
    if (publication == null) { $log.error(`Invalid App ID: ${receipt.bid}`); return }
    if (publication.itunesSharedSecret !== password) { $log.error('Invalid Shared Secret'); return }

    // Load and validate subscription
    const subscription = await ReaderSubscription.findOne({ where: { publicationId: publication.id, orderId: receipt.originalTransactionId } })
    if (subscription == null) { $log.error(`Subscription not found: ${receipt.originalTransactionId}`); return }

    // Update subscription
    subscription.endDate = receipt.expiresDate ? new Date(parseInt(receipt.expiresDate, 10)) : null
    subscription.verified = true
    subscription.receipt = generateLegacyReceiptData(unifiedReceipt.status, receipt, encodedReceipt)
    await subscription.save()

    $log.info('[AppStoreWebhookController] Updated Reader Subscription', receipt.originalTransactionId)
  }
}
