import { Publication, RestHook } from '@magloft/db-client'
import { BodyParams, Controller, Delete, Get, PathParams, Post, Req } from '@tsed/common'
import { NotFound } from '@tsed/exceptions'
import { Allow, Example, Required, Returns } from '@tsed/schema'
import { RestHookEntity, RestHookEvent } from '../../entities/RestHookEntity'
import { Auth, AuthRequest } from '../../middlewares/AuthMiddleware'

export class CreateRestHookParams {
  @Required() @Allow(['reader:signup', 'reader:signin', 'reader:signout', 'reader:signup']) @Example('reader:signup') event: RestHookEvent
  @Required() @Example('https://hooks.website.com/1234') targetUrl: string
}

@Controller('/rest')
export class RestHookController {
  @Get('/:publicationId/:id')
  @Auth()
  @Returns(200, RestHookEntity)
  async get(@Req() { user }: AuthRequest, @PathParams('publicationId') publicationId: number, @PathParams('id') id: number): Promise<RestHookEntity> {
    // Retrieve Publication
    const publication = await Publication.findOne({ where: { id: publicationId, userId: user.id } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }

    // Find Rest Hook
    const restHook = await RestHook.findOne({ where: { publicationId, id } })
    if (restHook == null) { throw new NotFound(`Rest Hook ${id} not found`) }

    // Return Rest Hook
    return RestHookEntity.from(restHook)
  }

  @Post('/:publicationId')
  @Auth()
  @Returns(200, RestHookEntity)
  async post(@Req() { user }: AuthRequest, @BodyParams(CreateRestHookParams) { event, targetUrl }: CreateRestHookParams, @PathParams('publicationId') publicationId: number): Promise<RestHookEntity> {
    // Retrieve Publication
    const publication = await Publication.findOne({ where: { id: publicationId, userId: user.id } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }

    // Find or Create Rest Hook
    const [restHook] = await RestHook.findOrCreate({ where: { event, targetUrl, publicationId } })

    // Return Rest Hook
    return RestHookEntity.from(restHook)
  }

  @Delete('/:publicationId/:id')
  @Auth()
  @Returns(200, RestHookEntity)
  async delete(@Req() { user }: AuthRequest, @PathParams('publicationId') publicationId: number, @PathParams('id') id: number): Promise<RestHookEntity> {
    // Retrieve Publication
    const publication = await Publication.findOne({ where: { id: publicationId, userId: user.id } })
    if (publication == null) { throw new NotFound(`Publication ${publicationId} not found`) }

    // Find Rest Hook
    const restHook = await RestHook.findOne({ where: { publicationId, id } })
    if (restHook == null) { throw new NotFound(`Rest Hook ${id} not found`) }

    // Delete Rest Hook
    await restHook.destroy()

    // Return Rest Hook
    return RestHookEntity.from(restHook)
  }
}
