import { FlexpdfInvoice, User } from '@magloft/db-client'
import { $log, Controller, Post, Req } from '@tsed/common'
import { Exception, NotFound, UnprocessableEntity } from '@tsed/exceptions'
import Stripe from 'stripe'
import { ReqWithStripeEvent, VerifyStripe } from '../../middlewares/VerifyStripeMiddleware'

export type InvoiceStatus = 'draft' | 'open' | 'paid' | 'uncollectible' | 'void'

export interface InvoiceMetadata {
  [key: string]: number | string | boolean
}

export interface InvoiceStatusTransitions {
  finalizedAt: number
  markedUncollectibleAt: number
  paidAt: number
  voidedAt: number
}

export const INVOICE_EVENT_TYPES = ['invoice.updated', 'invoice.marked_uncollectible', 'invoice.payment_failed', 'invoice.payment_succeeded', 'invoice.voided']

@Controller('/stripe-invoice')
export class StripeInvoiceController {
  @Post()
  @VerifyStripe('default')
  async post(@Req() { event }: ReqWithStripeEvent): Promise<void> {
    if (!INVOICE_EVENT_TYPES.includes(event.type)) { return }

    const invoice: Stripe.Invoice = event.data.object as Stripe.Invoice
    if (invoice.metadata.type !== 'flexpdf-credit') { return }

    // Validate User
    const user = await User.findOne({ where: { stripeId: String(invoice.customer) } })
    if (user == null) { return this.logError(new NotFound(`User with Stripe ID ${invoice.customer} not found`)) }

    // Validate Invoice
    const flexpdfInvoice = await FlexpdfInvoice.findOne({ where: { source: 'stripe', sourceInvoiceId: invoice.id } })
    if (flexpdfInvoice == null) { return this.logError(new NotFound(`FlexPDF Invoice #${invoice.id} not found`)) }
    if (flexpdfInvoice.paidAt || flexpdfInvoice.cancelledAt) { return this.logError(new UnprocessableEntity(`FlexPDF Invoice #${invoice.id} already processed`)) }

    // Update Invoice
    if (invoice.status === 'paid') {
      await Promise.all([
        flexpdfInvoice.update({ sourceInvoice: invoice, paidAt: new Date(invoice.status_transitions.paid_at * 1000) }),
        user.update({ flexpdfCredits: user.flexpdfCredits + flexpdfInvoice.credits })
      ])
    } else if (invoice.status === 'void') {
      await flexpdfInvoice.update({ sourceInvoice: invoice, cancelledAt: new Date(invoice.status_transitions.voided_at * 1000) })
    } else {
      await flexpdfInvoice.update({ sourceInvoice: invoice })
    }
  }

  private logError(error: Exception) {
    $log.error(`webhooks/stripe-invoice: [${error.status}] ${error.message}`)
  }
}
