import { Env } from '@tsed/core'
import { resolve } from 'path'
import { AppControllers } from './api/app'
import { FlexpdfControllers } from './api/flexpdf'
import { PortalControllers } from './api/portal'
import { PwaControllers } from './api/pwa'
import { WebhooksControllers } from './api/webhooks'
import { ArticleResolver } from './graphql/articles/ArticleResolver'
import { IssueResolver } from './graphql/issues/IssueResolver'
import { GraphQLQueryContextPlugin } from './graphql/plugins/GraphQLQueryContextPlugin'
import { PwaResolver } from './graphql/pwas/PwaResolver'
import { ReaderResolver } from './graphql/readers/ReaderResolver'
import { UserResolver } from './graphql/users/UserResolver'
import { AuthRequest } from './middlewares/AuthMiddleware'
import { AuthConfiguration } from './services/AuthService'

const rootDir = resolve(__dirname)
const env = process.env.KMS_ENV === 'production' ? Env.PROD : Env.DEV

export interface CorsConfiguration {
  whitelist: string[]
}

export interface Configuration extends TsED.Configuration {
  auth: AuthConfiguration
  cors: CorsConfiguration
}

export const config: Partial<Configuration> = {
  env,
  rootDir,
  httpsPort: false,
  ajv: {
    errorFormatter: (error) => `At ${error.modelName}${error.dataPath}, value '${error.data}' ${error.message}`,
    verbose: false
  },
  mount: {
    '/portal': PortalControllers,
    '/app': AppControllers,
    '/webhooks': WebhooksControllers,
    '/webhooks/v2': WebhooksControllers,
    '/flexpdf': FlexpdfControllers,
    '/pwa': PwaControllers
  },
  componentsScan: [
    '${rootDir}/graphql/**/*{.ts,.js}',
    '${rootDir}/mappers/**/*{.ts,.js}',
    '${rootDir}/middlewares/*{.ts,.js}',
    '${rootDir}/services/*{.ts,.js}'
  ],
  graphql: {
    magloft: {
      path: '/graphql',
      serverConfig: {
        playground: true,
        context: ({ req }: { req: AuthRequest }) => ({ user: req.user, role: req.role }),
        plugins: [GraphQLQueryContextPlugin]

      },
      buildSchemaOptions: {
        dateScalarMode: 'timestamp',
        resolvers: [ArticleResolver, IssueResolver, ReaderResolver, PwaResolver, UserResolver],
        authChecker: ({ context }, roles) => {
          if (roles.includes('user') && !context.user) { return false }
          if (roles.includes('admin') && context.user?.role !== 'admin') { return false }
          return true
        },
        validate: false
      }
    }
  },
  acceptMimes: ['application/json'],
  routers: {
    mergeParams: false,
    strict: false,
    caseSensitive: false
  },
  converter: {
    additionalProperties: 'ignore'
  },
  swagger: [{
    path: '',
    specPath: '${rootDir}/assets/swagger/swagger.base.json',
    cssPath: '${rootDir}/assets/swagger/theme-magloft.css'
  }]
}
