import { Article, Issue } from '@magloft/db-client'
import { CollectionOf, Description, Example, Property, Required, Schema } from '@tsed/schema'
import { Entity } from '../lib/Entity'

export interface ArticleWithSearch extends Article {
  score: number
  Issue: Issue
}

export enum ArticleSource { TYPELOFT = 'typeloft' }
export enum ArticleUnlockType { FREE = 'free', SUBSCRIBE = 'subscribe', OPTIN = 'optin' }

@Schema({ title: 'Article' })
export class ArticleEntity extends Entity {
  @Property()
  @Description('Unique ID')
  @Example(1)
  id: number

  @Required()
  @Description('Title')
  @Example('Sample Article')
  title: string

  @Property()
  @Description('HTML')
  @Example('<page-widget></page-widget>')
  html: string

  @Property()
  @Description('Unlock Type')
  @Example(ArticleUnlockType.FREE)
  unlockType: string

  @Property()
  @Description('Source')
  @Example(ArticleSource.TYPELOFT)
  source: string

  @Property()
  @Description('Position')
  @Example(0)
  position: number

  @Property()
  @Description('Issue ID')
  @Example(1)
  issueId: number

  @Property()
  @Description('Cover')
  @Example('defaults/images/cover.jpg')
  cover: string

  @CollectionOf(Number)
  @Description('Category IDs')
  @Example([1, 2, 3])
  categoryIds: number[]
}

@Schema({ title: 'Article Search' })
export class ArticleSearchEntity extends Entity {
  @Property()
  @Description('Unique ID')
  @Example(1)
  id: number

  @Required()
  @Description('Title')
  @Example('Sample Article')
  title: string

  @Property()
  @Description('Cover')
  @Example('https://cdn.magloft.com/images/sample.jpg')
  cover: string | null

  @Property()
  @Description('Issue ID')
  @Example(1)
  issueId: number | null

  @Property()
  @Description('Issue Name')
  @Example('issue.name.1234')
  issueName: string | null

  @Property()
  @Description('Issue Title')
  @Example('Test Issue')
  issueTitle: string | null

  @Property()
  @Description('Issue Cover')
  @Example('https://cdn.magloft.com/images/sample.jpg')
  issueCover: string | null

  @Property()
  @Description('Score')
  @Example(40.0)
  score: number

  @Property()
  @Description('Match')
  @Example('Extracted text match')
  match: string

  constructor(model?: ArticleWithSearch, attributes?: object) {
    super(model, attributes)
    this.issueTitle = model.Issue ? model.Issue.title : null
    this.issueName = model.Issue ? model.Issue.name : null
    this.issueCover = model.Issue?.cover ? `https://cdn.magloft.com/${model.Issue.cover}` : null
  }
}
