import * as db from '@magloft/db-client'
import { col, fn, Op, where, WhereOptions } from '@magloft/db-client'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { Article, ArticleList } from '../../articles/Article'
import { GetByIdArgs, GetListArgs } from './AdminArticleInput'

@DataSourceService()
export class AdminArticleDataSource extends DataSource {
  public async getById({ id }: GetByIdArgs, user: db.User): Promise<Article> {
    // Validate Article
    const article = await db.Article.findById(id)
    if (!article) { throw new Error('Article not found') }

    // Return Article
    return new Article(article)
  }

  async getList({ publicationId, offset, limit, orderBy, orderDir, ids, source, unlockType, categoryId, categoryIds, searchText }: GetListArgs, user: db.User): Promise<ArticleList> {
    const publicationCount = await db.Publication.count({ where: { id: publicationId, userId: user.id } })
    if (publicationCount === 0) { throw new Error('Publication not found') }

    // Prepare Options
    const options: WhereOptions[] = [{ publicationId }]
    if (ids) { options.push({ id: { [Op.in]: ids } }) }
    if (source) { options.push({ source }) }
    if (unlockType) { options.push({ unlockType }) }
    if (categoryId) { options.push(where(fn('JSON_CONTAINS', col('Article.category_ids'), JSON.stringify([categoryId])), '=', '1')) }
    if (categoryIds && categoryIds.length > 0) { options.push(where(fn('JSON_CONTAINS', col('Article.category_ids'), JSON.stringify(categoryIds)), '=', '1')) }
    if (searchText) { options.push({ title: { [Op.like]: `%${searchText}%` } }) }

    // Load Articles
    const articles = await db.Article.findAll({ offset, limit, where: { [Op.and]: options }, order: [[orderBy, orderDir]] })

    // Return Article List
    const list = new ArticleList()
    list.offset = offset
    list.limit = limit
    list.records = articles.map((article) => new Article(article))
    return list
  }
}
