import * as db from '@magloft/db-client'
import { col, fn, Op, where, WhereOptions } from '@magloft/db-client'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { Article, ArticleList } from './Article'
import { GetByIdArgs, GetBySlugArgs, GetListArgs } from './ArticleInput'

@DataSourceService()
export class ArticleDataSource extends DataSource {
  public async getById({ id }: GetByIdArgs): Promise<Article> {
    // Validate Article
    const article = await db.Article.findById(id)
    if (!article) { throw new Error('Article not found') }

    // Return Article
    return new Article(article)
  }

  public async getBySlug({ slug }: GetBySlugArgs): Promise<Article> {
    // Validate Article
    const article = /^article-([0-9]+)$/.test(slug)
      ? await db.Article.findOne({ where: { slug: { [Op.eq]: null }, id: parseInt(slug.split('-')[1], 10) } })
      : await db.Article.findOne({ where: { slug } })
    if (!article) { throw new Error('Article not found') }

    // Return Article
    return new Article(article)
  }

  async getList({ publicationId, offset, limit, orderBy, orderDir, ids, source, unlockType, categoryId, categoryIds, searchText }: GetListArgs): Promise<ArticleList> {
    // Prepare Options
    const options: WhereOptions[] = [{ publicationId, date: { [Op.ne]: null }, html: { [Op.ne]: null }, standalone: true }]
    if (ids) { options.push({ id: { [Op.in]: ids } }) }
    if (source) { options.push({ source }) }
    if (unlockType) { options.push({ unlockType }) }
    if (categoryId) { options.push(where(fn('JSON_CONTAINS', col('Article.category_ids'), JSON.stringify([categoryId])), '=', '1')) }
    if (categoryIds && categoryIds.length > 0) { options.push(where(fn('JSON_CONTAINS', col('Article.category_ids'), JSON.stringify(categoryIds)), '=', '1')) }
    if (searchText) { options.push({ title: { [Op.like]: `%${searchText}%` } }) }

    // Load Articles
    const articles = await db.Article.findAll({ offset, limit, where: { [Op.and]: options }, order: [[orderBy, orderDir]] })

    // Return Article List
    const list = new ArticleList()
    list.offset = offset
    list.limit = limit
    list.records = articles.map((article) => new Article(article))
    return list
  }
}
