import * as db from '@magloft/db-client'
import { col, fn, Op, where, WhereOptions } from '@magloft/db-client'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { Issue, IssueList } from './Issue'
import { GetByIdArgs, GetBySlugArgs, GetListArgs } from './IssueInput'

@DataSourceService()
export class IssueDataSource extends DataSource {
  public async getById({ id }: GetByIdArgs): Promise<Issue> {
    const issue = await db.Issue.findOne({ where: { id, slug: { [Op.ne]: null } }, include: [db.Issue.associations['Articles']] })
    if (!issue) { throw new Error('Issue not found') }

    // Return Issue
    return new Issue(issue)
  }

  public async getBySlug({ slug }: GetBySlugArgs): Promise<Issue> {
    const issue = await db.Issue.findOne({ where: { slug }, include: [db.Issue.associations['Articles']] })
    if (!issue) { throw new Error('Issue not found') }

    // Return Issue
    return new Issue(issue)
  }

  async getList({ publicationId, offset, limit, orderBy, orderDir, ids, categoryId, categoryIds, searchText, source, unlockType }: GetListArgs): Promise<IssueList> {
    // Prepare Options
    const options: WhereOptions[] = [{ publicationId, date: { [Op.ne]: null }, slug: { [Op.ne]: null } }]
    if (ids) { options.push({ id: { [Op.in]: ids } }) }
    if (source) { options.push({ source }) }
    if (unlockType) { options.push({ unlockType }) }
    if (categoryId) { options.push(where(fn('JSON_CONTAINS', col('Issue.category_ids'), JSON.stringify([categoryId])), '=', '1')) }
    if (categoryIds && categoryIds.length > 0) { options.push(where(fn('JSON_CONTAINS', col('Issue.category_ids'), JSON.stringify(categoryIds)), '=', '1')) }
    if (searchText) { options.push({ title: { [Op.like]: `%${searchText}%` } }) }

    // Load Issues
    const include = [{ model: db.Article, as: 'Articles', required: false, attributes: ['id', 'slug'] }]
    const issues = await db.Issue.findAll({ offset, limit, where: { [Op.and]: options }, order: [[orderBy, orderDir]], include })

    // Return Issue List
    const list = new IssueList()
    list.offset = offset
    list.limit = limit
    list.records = issues.map((issue) => new Issue(issue))
    return list
  }
}
