import * as db from '@magloft/db-client'
import { Field, Float, Int, ObjectType } from 'type-graphql'
import { BaseSchema } from '../../lib/graphql/BaseSchema'
import { Article } from '../articles/Article'
import { Issue } from '../issues/Issue'

export type ReaderPurchaseWithIncludes = db.ReaderPurchase & {
  Issue?: db.Issue
  Article?: db.Article
}

export type ReaderSubscriptionWithIncludes = db.ReaderSubscription & {
  AppSubscription: db.AppSubscription
}

@ObjectType()
export class Checkout {
  @Field() public sessionId: string
}

@ObjectType()
export class BillingPortal {
  @Field() public url: string
}

@ObjectType()
export class AppSubscription extends BaseSchema<db.AppSubscription> {
  @Field(type => Int) public id: number
  @Field(type => Int, { nullable: true }) public publicationId?: number
  @Field() public title: string
  @Field({ nullable: true }) public description?: string
  @Field(type => Float) public price: number
  @Field() public interval: string
  @Field(type => Int) public intervalCount: number
  @Field() public unlockType: string

  constructor(appSubscription: db.AppSubscription) {
    super(appSubscription)
    const duration = appSubscription.duration.split('.')
    this.interval = duration[0]
    this.intervalCount = parseInt(duration[1], 10)
    this.price = this.model.price - 0.01
  }
}

@ObjectType()
export class ReaderSubscription extends BaseSchema<db.ReaderSubscription> {
  @Field(type => Int) public id: number
  @Field({ nullable: true }) public orderId?: string
  @Field({ nullable: true }) public source?: string
  @Field({ nullable: true }) public userId?: string
  @Field(type => Int, { nullable: true }) public publicationId?: number
  @Field(type => Int, { nullable: true }) public readerId?: number
  @Field({ nullable: true }) public active?: boolean
  @Field({ nullable: true }) public startDate?: Date
  @Field({ nullable: true }) public endDate?: Date
  @Field(type => Int, { nullable: true }) public appSubscriptionId?: number
  @Field({ nullable: true }) public verified?: boolean
  @Field(type => AppSubscription) public appSubscription?: AppSubscription
  @Field() public createdAt: Date

  constructor(subscription: ReaderSubscriptionWithIncludes) {
    super(subscription)
    this.appSubscription = new AppSubscription(subscription.AppSubscription)
  }
}

@ObjectType()
export class ReaderPurchase extends BaseSchema<db.ReaderPurchase> {
  @Field(type => Int) public id: number
  @Field(type => Int, { nullable: true }) public readerId?: number
  @Field(type => Int, { nullable: true }) public publicationId?: number
  @Field(type => Int, { nullable: true }) public issueId?: number
  @Field({ nullable: true }) public orderId?: string
  @Field({ nullable: true }) public source?: string
  @Field({ nullable: true }) public userId?: string
  @Field(type => Int, { nullable: true }) public articleId?: number
  @Field({ nullable: true }) public verified?: boolean
  @Field(type => Issue, { nullable: true }) public issue?: Issue
  @Field(type => Article, { nullable: true }) public article?: Article
  @Field() public createdAt: Date

  constructor(purchase: ReaderPurchaseWithIncludes) {
    super(purchase)
    if (purchase.Issue) { this.issue = new Issue(purchase.Issue) }
    if (purchase.Article) { this.article = new Article(purchase.Article) }
  }
}
