import * as db from '@magloft/db-client'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { ApolloError } from 'apollo-server-express'
import { PaymentScope } from '../../lib/payments/PaymentProvider'
import { PaymentService } from '../../services/PaymentService'
import { BillingPortal, Checkout } from './Purchase'
import { CreateBillingPortalSessionArgs, CreateCheckoutSessionArgs } from './PurchaseInput'

@DataSourceService()
export class PurchaseDataSource extends DataSource {
  constructor(private payment?: PaymentService) { super() }

  async createCheckoutSession({ pwaId, readerId, kind, id, extra, providerIdentifier, returnUrl }: CreateCheckoutSessionArgs): Promise<Checkout> {
    // Retrieve PWA
    const pwa = await db.Pwa.findById(pwaId)
    if (!pwa) { throw new Error('PWA not found') }

    // Retrieve Publication
    const publication = await db.Publication.findById(pwa.publicationId)
    if (!publication) { throw new ApolloError('Publication not found', 'PUBLICATION_NOT_FOUND') }

    // Create Reader
    const reader = await db.Reader.findOne({ where: { publicationId: pwa.publicationId, id: readerId } })
    if (!reader) { throw new ApolloError('Reader not found', 'READER_NOT_FOUND') }

    // Retrieve Provider
    const provider = this.payment.getProvider(pwa, providerIdentifier)
    const scope: PaymentScope = { publication, kind, id }

    // Validate Access to Issue
    await this.payment.ensurePurchaseAvailable(reader, scope)

    // Sync Product
    const product = await this.payment.syncScope(provider, scope)

    // Create Checkout Session
    const session = await provider.createCheckoutSession({ pwa, reader, product, returnUrl, extra })

    // Return Session ID
    const checkout = new Checkout()
    checkout.sessionId = session.id
    return checkout
  }

  async createBillingPortalSession({ pwaId, readerId, providerIdentifier, returnUrl }: CreateBillingPortalSessionArgs): Promise<BillingPortal> {
    // Retrieve PWA
    const pwa = await db.Pwa.findById(pwaId)
    if (!pwa) { throw new Error('PWA not found') }

    // Create Reader
    const reader = await db.Reader.findOne({ where: { publicationId: pwa.publicationId, id: readerId } })
    if (!reader) { throw new Error('Reader not found') }

    // Retrieve Provider
    const provider = this.payment.getProvider(pwa, providerIdentifier)

    // Create Billing Portal Session
    const session = await provider.createBillingPortalSession({ reader, returnUrl })

    // Return Session ID
    const billingPortal = new BillingPortal()
    billingPortal.url = session.url
    return billingPortal
  }
}
