import * as db from '@magloft/db-client'
import { UnprocessableEntity } from '@tsed/exceptions'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { validate } from 'class-validator'
import { ReaderEntity } from '../../entities/ReaderEntity'
import { AuthService } from '../../services/AuthService'
import { RestHookService } from '../../services/RestHookService'
import { Reader } from './Reader'
import { GetByTokenArgs, RefreshArgs, SignInArgs, SignOutArgs, SignUpArgs, UpdateArgs } from './ReaderInput'

const include = [{
  model: db.ReaderSubscription,
  as: 'ReaderSubscriptions',
  required: false,
  where: { active: true },
  include: [
    { model: db.AppSubscription, as: 'AppSubscription', required: true }
  ]
}, {
  model: db.ReaderPurchase,
  as: 'ReaderPurchases',
  required: false,
  include: [
    { model: db.Issue, as: 'Issue', required: false },
    { model: db.Article, as: 'Article', required: false }
  ]
}]

export type PublicationWithRestHooks = db.Publication & { RestHooks: db.RestHook[] }

@DataSourceService()
export class ReaderDataSource extends DataSource {
  constructor(private auth: AuthService, private restHook: RestHookService) { super() }

  public async getByToken({ token }: GetByTokenArgs): Promise<Reader> {
    // Validate Token & Reader
    const reader = await this.auth.decrypt<db.Reader>(token, 'reader', (id) => db.Reader.findOne({ where: { id }, include }))
    if (!reader) { throw new Error('Reader not found') }

    // Return Reader
    return new Reader(reader as db.Reader, token)
  }

  public async signIn(args: SignInArgs): Promise<Reader> {
    // Validate Input
    const validationErrors = await validate(args, { skipMissingProperties: true })
    if (validationErrors && validationErrors.length > 0) { throw new UnprocessableEntity('The submission contains invalid data') }
    const { email, password, publicationId } = args

    // Validate Publication
    const options = { include: [{ model: db.RestHook, as: 'RestHooks', required: false, where: { event: 'reader:signin' } }] }
    const publication = await db.Publication.findById(publicationId, options) as PublicationWithRestHooks
    if (!publication) { throw new Error('Publication not found') }

    // Validate Reader
    const reader = await db.Reader.findOne({ where: { email, publicationId: publication.id }, include })
    if (!reader) { throw new Error('Reader not found') }

    // Validate Password
    if (!this.auth.validatePassword(reader.encryptedPassword, password)) { throw new Error('Incorrect Password') }

    // Trigger Rest Hooks
    this.restHook.notify<ReaderEntity>(publication.RestHooks, ReaderEntity.from(reader))

    // Return Reader
    return new Reader(reader, this.auth.encrypt(reader))
  }

  public async refresh(args: RefreshArgs): Promise<Reader> {
    // Validate Input
    const validationErrors = await validate(args, { skipMissingProperties: true })
    if (validationErrors && validationErrors.length > 0) { throw new UnprocessableEntity('The submission contains invalid data') }
    const { token } = args

    // Validate Token & Reader
    const reader = await this.auth.decrypt<db.Reader>(token, 'reader', (id) => db.Reader.findOne({ where: { id }, include }))
    if (!reader) { throw new Error('Reader not found') }

    // Return Reader
    return new Reader(reader, this.auth.encrypt(reader))
  }

  public async signUp(args: SignUpArgs): Promise<Reader> {
    // Validate Input
    const validationErrors = await validate(args, { skipMissingProperties: true })
    if (validationErrors && validationErrors.length > 0) { throw new UnprocessableEntity('The submission contains invalid data') }
    const { email, password, publicationId, firstname, lastname, newsletter, optinId } = args

    // Validate Publication
    const options = { include: [{ model: db.RestHook, as: 'RestHooks', required: false, where: { event: 'reader:signup' } }] }
    const publication = await db.Publication.findById(publicationId, options) as PublicationWithRestHooks
    if (!publication) { throw new Error('Publication not found') }

    // Validate Existing Reader
    const existingReader = await db.Reader.findOne({ where: { email, publicationId: publication.id } })
    if (existingReader) { throw new Error('Reader already Exist!') }

    // Create Reader
    const createdAt = new Date()
    const reader = await publication.createReader({
      email,
      publicationId,
      createdAt,
      encryptedPassword: this.auth.encryptPassword(password ?? this.auth.getDefaultPassword(email, createdAt)),
      newsletter: newsletter ?? false,
      name: firstname,
      lastname,
      source: optinId ? 'optin' : 'default'
    })
    if (!reader.id) { throw new Error('Cannot create Reader') }

    // Trigger Rest Hooks
    this.restHook.notify<ReaderEntity>(publication.RestHooks, ReaderEntity.from(reader, { password }))

    // Return Reader
    return new Reader(reader, this.auth.encrypt(reader))
  }

  public async update(args: UpdateArgs): Promise<Reader> {
    // Validate Input
    const validationErrors = await validate(args, { skipMissingProperties: true })
    if (validationErrors && validationErrors.length > 0) { throw new UnprocessableEntity('The submission contains invalid data') }
    const { token, email, password, firstname, lastname, newsletter } = args

    // Validate Reader
    const reader = await this.auth.decrypt<db.Reader>(token, 'reader', (id) => db.Reader.findOne({ where: { id }, include }))
    if (!reader) { throw new Error('Reader not found') }

    // Fetch Publication
    const options = { include: [{ model: db.RestHook, as: 'RestHooks', required: false, where: { event: 'reader:update' } }] }
    const publication = await db.Publication.findById(reader.publicationId, options) as PublicationWithRestHooks
    if (!publication) { throw new Error('Publication not found') }

    // Update Reader
    if (email !== undefined) { reader.email = email }
    if (password !== undefined && !this.auth.validatePassword(reader.encryptedPassword, password)) {
      reader.encryptedPassword = this.auth.encryptPassword(password)
    }
    if (firstname !== undefined) { reader.firstname = firstname }
    if (lastname !== undefined) { reader.lastname = lastname }
    if (newsletter !== undefined) { reader.newsletter = newsletter }
    await reader.save()

    // Trigger Rest Hooks
    this.restHook.notify<ReaderEntity>(publication.RestHooks, ReaderEntity.from(reader))

    // Return Reader
    return new Reader(reader, this.auth.encrypt(reader))
  }

  public async signOut({ token }: SignOutArgs): Promise<Reader> {
    // Validate Reader
    const reader = await this.auth.decrypt<db.Reader>(token, 'reader')
    if (!reader) { throw new Error('Reader not found') }

    // Fetch Publication
    const options = { include: [{ model: db.RestHook, as: 'RestHooks', required: false, where: { event: 'reader:signout' } }] }
    const publication = await db.Publication.findById(reader.publicationId, options) as PublicationWithRestHooks
    if (!publication) { throw new Error('Publication not found') }

    // Trigger Rest Hooks
    this.restHook.notify<ReaderEntity>(publication.RestHooks, ReaderEntity.from(reader))

    // Return Reader
    return new Reader(reader)
  }
}
