import * as db from '@magloft/db-client'
import { UnprocessableEntity } from '@tsed/exceptions'
import { DataSourceService } from '@tsed/graphql'
import { DataSource } from 'apollo-datasource'
import { validate } from 'class-validator'
import { AuthService } from '../../services/AuthService'
import { User } from './User'
import { GetByTokenArgs, SignInArgs, SignOutArgs } from './UserInput'

@DataSourceService()
export class UserDataSource extends DataSource {
  constructor(private auth: AuthService) { super() }

  public async getByToken({ token }: GetByTokenArgs): Promise<User> {
    // Validate Token & User
    const user = await this.auth.decrypt<db.User>(token, 'user')
    if (!user) { throw new Error('User not found') }

    // Return User
    return new User(user, token)
  }

  public async signIn(args: SignInArgs): Promise<User> {
    // Validate Input
    const validationErrors = await validate(args, { skipMissingProperties: true })
    if (validationErrors && validationErrors.length > 0) { throw new UnprocessableEntity('The submission contains invalid data') }
    const { email, password } = args

    // Validate User
    const user = await db.User.findOne({ where: { email } })
    if (!user) { throw new Error('User not found') }

    // Validate Password
    if (!this.auth.validatePassword(user.encryptedPassword, password)) { throw new Error('Incorrect Password') }

    // Return User
    return new User(user, this.auth.encrypt(user))
  }

  public async signOut({ token }: SignOutArgs): Promise<User> {
    // Validate User
    const user = await this.auth.decrypt<db.User>(token, 'user')
    if (!user) { throw new Error('User not found') }

    // Return Reader
    return new User(user)
  }
}
