import { BaseModel } from '@magloft/db-client/lib/models/base/Model.base'
import { EndpointInfo, IMiddleware, Middleware, Next, QueryParams, Req, UseBefore } from '@tsed/common'
import { applyDecorators, StoreSet } from '@tsed/core'
import { Entity } from '../Entity'
import { ListEntity } from '../ListEntity'

export interface Pagination {
  offset: number
  limit: number
  orderBy: string
  orderDir: string
}

export type PaginateRequest = Express.Request & Pagination & {
  paginate?: (totalAsync: Promise<number>, modelsAsync: Promise<BaseModel[]>) => Promise<ListEntity<Entity>>
}

export interface PaginateMiddlewareConfig {
  useType: typeof Entity
  defaults: Pagination
}

@Middleware()
export class PaginateMiddleware implements IMiddleware {
  public async use(
    @Next() next: Next,
    @Req() request: PaginateRequest,
    @EndpointInfo() endpoint: EndpointInfo,
    @QueryParams('offset', Number) offset?: number,
    @QueryParams('limit', Number) limit?: number,
    @QueryParams('orderBy', String) orderBy?: string,
    @QueryParams('orderDir', String) orderDir?: string
  ) {
    const { defaults, useType }: PaginateMiddlewareConfig = endpoint.get(PaginateMiddleware)
    request.offset = offset !== undefined ? offset : defaults.offset
    request.limit = limit !== undefined ? limit : defaults.limit
    request.orderBy = orderBy !== undefined ? orderBy : defaults.orderBy
    request.orderDir = orderDir !== undefined ? orderDir : defaults.orderDir
    request.paginate = async (totalAsync, modelsAsync) => {
      const [total, models] = await Promise.all([totalAsync, modelsAsync])
      const records = useType.map(models)
      return { total, offset: request.offset, limit: request.limit, records }
    }
    return next()
  }
}

export function Paginate(useType: typeof Entity, defaults: Partial<Pagination> = {}): Function {
  const config: PaginateMiddlewareConfig = { useType, defaults: { offset: 0, limit: 100, orderBy: 'id', orderDir: 'desc', ...defaults } }
  return applyDecorators(StoreSet(PaginateMiddleware, config), UseBefore(PaginateMiddleware))
}
