import { capitalCase, paramCase } from 'change-case'
import { Declaration, FontFace, parse } from 'css'

export interface Font {
  identifier: string
  title: string
  variants: FontVariant[]
}

export interface FontVariant {
  style: 'normal' | 'italic'
  weight: 'normal' | 'bold'
  urls: { [key: string]: string }
}

export function parseFontSrc(src: string): { [key: string]: string } {
  const matches = src.match(/url\([^)]+\)\sformat\([^)]+\)/g) ?? []
  return matches.reduce<{ [key: string]: string }>((obj, part) => {
    const format = part.match(/format\("?'?([^)"']+)"?'?\)/)
    const url = part.match(/url\("?([^)"]+)"?\)/)
    if (!url || !format) { return obj }
    obj[format[1]] = url[1]
    return obj
  }, {})
}

export function parseFonts(text: string) {
  const { stylesheet } = parse(text)
  const rules: FontFace[] = stylesheet.rules.filter((rule) => rule.type === 'font-face')
  return rules.reduce<Font[]>((arr, rule) => {
    if (!rule.declarations) { return arr }
    const fontFamilyDeclaration: Declaration = rule.declarations.find(({ property }: Declaration) => property === 'font-family')
    if (!fontFamilyDeclaration || !fontFamilyDeclaration.value) { return arr }
    const identifier = paramCase(fontFamilyDeclaration.value)
    const title = capitalCase(identifier)

    const font: Font = arr.find((item) => item.identifier === identifier) ?? { identifier, title, variants: [] }
    if (!arr.includes(font)) { arr.push(font) }

    const fontStyle: Declaration = rule.declarations.find(({ property }: Declaration) => property === 'font-style')
    const fontWeight: Declaration = rule.declarations.find(({ property }: Declaration) => property === 'font-weight')
    const src: Declaration = rule.declarations.find(({ property }: Declaration) => property === 'src')
    if (!src || !src.value) { return arr }

    const variant: FontVariant = {
      style: (fontStyle?.value ?? 'normal') as 'normal' | 'italic',
      weight: ['400', 'normal'].includes(fontWeight.value) ? 'normal' : 'bold',
      urls: parseFontSrc(src.value)
    }

    font.variants.push(variant)
    return arr
  }, [])
}
