import { Pwa, Reader } from '@magloft/db-client'
import { Configuration } from '@tsed/di'
import Stripe from 'stripe'
import { StripeConfiguration } from '../../services/StripeService'
import { BillingPortalSession, CheckoutSession, CreateBillingPortalParams, CreateCheckoutSessionParams, CreateProductParams, PaymentProvider, Price, PriceInterval, Product, Unlock } from './PaymentProvider'

function mapPrice({ id, currency, unit_amount, recurring, type }: Stripe.Price): Price {
  const interval = type === 'recurring' ? recurring.interval as PriceInterval : PriceInterval.ONCE
  const price: Price = { id, currency, amount: unit_amount / 100, interval }
  if (type === 'recurring') { price.intervalCount = recurring.interval_count }
  return price
}

function mapProduct({ id, name, description, images, metadata }: Stripe.Product, price: Stripe.Price): Product {
  return {
    id,
    provider: 'stripe',
    name,
    description,
    images,
    price: mapPrice(price),
    unlocks: mapMetadataToUnlocks(metadata)
  }
}

function mapCheckoutSession({ id, subscription, payment_intent }: Stripe.Checkout.Session, product: Product, readerId: number, publicationId: number, extra?: number): CheckoutSession {
  return {
    id,
    provider: 'stripe',
    product,
    readerId,
    publicationId,
    orderId: (subscription as string) ?? (payment_intent as string),
    extra
  }
}

function mapBillingPortalSession({ id, url }: Stripe.BillingPortal.Session, readerId: number): BillingPortalSession {
  return { id, provider: 'stripe', url, readerId }
}

function mapUnlocksToMetadata(unlocks: Unlock[]): Stripe.MetadataParam {
  const result = unlocks.map(({ kind, id }) => `${kind}:${id}`).join(',')
  return { unlocks: result }
}

function mapMetadataToUnlocks(metadata: Stripe.MetadataParam): Unlock[] {
  const value: string = metadata.unlocks ? String(metadata.unlocks) : ''
  return value.split(',').map<Unlock>((entry) => {
    const args = entry.split(':')
    return { kind: args[0], id: parseInt(args[1], 10) } as Unlock
  })
}

function generateLookupKey(id: string, price: Price) {
  return (price.interval === PriceInterval.ONCE)
    ? `${id}-${price.interval}`
    : `${id}-${price.interval}-${price.intervalCount}`
}

export class StripePaymentProvider implements PaymentProvider {
  private client: Stripe
  private options: Stripe.RequestOptions

  constructor(public configuration: Configuration, public pwa: Pwa) {
    const { secretKey } = this.configuration.get<StripeConfiguration>('stripe')
    this.client = new Stripe(secretKey, { apiVersion: '2020-08-27' })
    this.options = { stripeAccount: pwa.config.payment.stripe.accountId }
  }

  async syncProduct(params: CreateProductParams): Promise<Product> {
    const product = await this.findOrCreateProduct(params)
    const price = await this.findOrCreatePrice(product, params.price)
    return mapProduct(product, price)
  }

  async createCheckoutSession({ pwa, product, reader, returnUrl, extra }: CreateCheckoutSessionParams): Promise<CheckoutSession> {
    const { price } = product
    const customer = await this.findOrCreateCustomer(reader)
    const session = await this.client.checkout.sessions.create({
      mode: (price.interval === PriceInterval.ONCE) ? 'payment' : 'subscription',
      client_reference_id: String(reader.id),
      customer: customer.id,
      payment_method_types: ['card'],
      line_items: [{ price: price.id, quantity: 1 }],
      cancel_url: `${returnUrl}?checkout=cancel`,
      success_url: `${returnUrl}?checkout=success`,
      metadata: { pwaId: pwa.id, readerId: reader.id, publicationId: pwa.publicationId, extra }
    }, this.options)
    return mapCheckoutSession(session, product, reader.id, pwa.publicationId, extra)
  }

  async createBillingPortalSession({ reader, returnUrl }: CreateBillingPortalParams): Promise<BillingPortalSession> {
    const session = await this.client.billingPortal.sessions.create({
      customer: reader.stripeId,
      return_url: returnUrl
    }, this.options)
    return mapBillingPortalSession(session, reader.id)
  }

  async retrieveCheckoutSession({ id }: Stripe.Checkout.Session): Promise<CheckoutSession> {
    const session = await this.client.checkout.sessions.retrieve(id, { expand: ['line_items', 'line_items.data.price.product'] }, this.options)
    const [item] = session.line_items.data
    const product = mapProduct(item.price.product as Stripe.Product, item.price)
    const readerId = parseInt(session.metadata.readerId, 10)
    const publicationId = parseInt(session.metadata.publicationId, 10)
    const extra = session.metadata.extra ? parseInt(session.metadata.extra, 10) : null
    return mapCheckoutSession(session, product, readerId, publicationId, extra)
  }

  private findOrCreateProduct({ id, name, description, images, unlocks }: CreateProductParams) {
    const metadata = mapUnlocksToMetadata(unlocks)
    return this.client.products.retrieve(id, this.options).catch(() => {
      return this.client.products.create({ id, name, description, images, metadata }, this.options)
    })
  }

  private async findOrCreatePrice({ id }: Stripe.Product, price: Price) {
    const lookupKey = generateLookupKey(id, price)
    const { data } = await this.client.prices.list({ product: id, lookup_keys: [lookupKey] }, this.options)
    if (data.length > 0) { return data[0] }
    const params: Stripe.PriceCreateParams = {
      currency: price.currency,
      unit_amount: price.amount * 100,
      product: id,
      lookup_key: lookupKey
    }
    if (price.interval !== PriceInterval.ONCE) {
      params.recurring = { interval: price.interval, interval_count: price.intervalCount }
    }
    // @TODO: Implement Trial Period
    return this.client.prices.create(params, this.options)
  }

  private async findOrCreateCustomer(reader: Reader) {
    if (reader.stripeId) { return this.client.customers.retrieve(reader.stripeId, this.options) }
    const customer = await this.client.customers.create({ email: reader.email, name: reader.name }, this.options)
    if (!reader.stripeId) {
      reader.stripeId = customer.id
      await reader.save()
    }
    return customer
  }
}
