import { User } from '@magloft/db-client'
import { CookiesParams, EndpointInfo, HeaderParams, IAuthOptions, IMiddleware, Middleware, Next, QueryParams, Req, UseAuth } from '@tsed/common'
import { applyDecorators } from '@tsed/core'
import { Forbidden, Unauthorized } from '@tsed/exceptions'
import { Returns, Security } from '@tsed/schema'
import { Request } from 'express'
import { AuthRole, AuthService } from '../services/AuthService'

export interface AuthOptions extends IAuthOptions {
  role: AuthRole
}

export interface AuthRequest extends Request {
  user?: User
  role?: AuthRole
}

@Middleware()
export class AuthMiddleware implements IMiddleware {
  constructor(private auth: AuthService) { }

  public async use(
    @Next() next: Next,
    @Req() request: AuthRequest,
    @QueryParams('token') queryToken?: string,
    @HeaderParams('x-magloft-token') headerToken?: string,
    @HeaderParams('x-magloft-accesstoken') legacyHeaderToken?: string,
    @CookiesParams() cookies?,
    @EndpointInfo() endpoint?: EndpointInfo
  ) {
    const options = (endpoint ? (endpoint.get(AuthMiddleware) || {}) : {}) as AuthOptions
    const token = queryToken || headerToken
    let user: User
    if (token) {
      user = await this.auth.decrypt<User>(token, 'user')
    } else if (legacyHeaderToken) {
      user = await this.auth.decryptLegacyHeaderToken(legacyHeaderToken)
    } else if (cookies?.[this.auth.cookieName]) {
      user = await this.auth.decryptCookie(cookies[this.auth.cookieName])
    }
    if (options.role !== 'anonymous' && !user) { return next(new Unauthorized('Unauthorized')) }
    if (options.role === 'admin' && user.role !== 'admin') { return next(new Forbidden('Forbidden')) }
    request.user = user
    request.role = options.role
    return next()
  }
}

export function Auth(role: AuthRole = 'user') {
  return applyDecorators(
    UseAuth(AuthMiddleware, { role }),
    Security('token'),
    Returns(401, { description: 'Unauthorized' }),
    Returns(403, { description: 'Forbidden' })
  )
}
