import { Category, close, connect, Op, Publication, Pwa, PwaOptin } from '@magloft/db-client'
import * as kms from '@magloft/kms'
import { PwaConfig } from '@magloft/schema/dist/pwa/config'
import { PwaDesign } from '@magloft/schema/dist/pwa/design'
import { PwaLayout } from '@magloft/schema/dist/pwa/layout'
import { capitalCase, paramCase } from 'change-case'
import { domain, publicationId } from './data/settings.json'

connect(kms.config('credentials/db/cloud-server'))

function cleanString(value: string) {
  value = value.replace(' - MagLoft', '')
  value = value.replace(/[\-]/g, ' ')
  value = value.replace(/[\"\'\(\)\!\?\,\:\.]/g, '')
  value = value.replace(/\s\s/g, ' ')
  value = value.replace(/\s\s/g, ' ')
  value = capitalCase(value)
  return value
}

async function main() {
  // Load Publciation
  const publication = await Publication.findById(publicationId)
  if (!publication) { throw new Error('MagLoft Publication does not exist') }

  // Prepare Categories
  console.info('~> Updating Categories')
  const [categoryA] = await Category.findOrCreate({ where: { publicationId, title: 'Business & Strategy' }, defaults: { color: '#ee0cda' } })
  const [categoryB] = await Category.findOrCreate({ where: { publicationId, title: 'Concept & Design' }, defaults: { color: '#ee0cda' } })
  const [categoryC] = await Category.findOrCreate({ where: { publicationId, title: 'Lifestyle & Entertainment' }, defaults: { color: '#ee0cda' } })
  const categories = [categoryA, categoryB, categoryC]

  // Prepare Articles
  console.info('~> Updating Articles')
  const articles = await publication.getArticles({ where: { date: { [Op.ne]: null }, html: { [Op.ne]: null } }, order: [['date', 'desc']] })
  for (const article of articles) {
    if (!article.keywordPhrase) { article.keywordPhrase = cleanString(article.title) }
    if (!article.slug) { article.slug = paramCase(article.keywordPhrase) }
    const category = categories[article.id % 3]
    article.categoryIds = [category.id]
    if (article.changed()) { await article.save() }
  }

  // Prepare Issues
  console.info('~> Updating Issues')
  const issues = await publication.getIssues({ where: { date: { [Op.ne]: null }, unlockType: 'free', source: 'typeloft2' }, order: [['date', 'desc']] })
  for (const issue of issues) {
    const category = categories[issue.id % 3]
    issue.categoryIds = [category.id]
    if (issue.changed()) { await issue.save() }
  }

  // Load PWA
  console.info('~> Updating PWA')
  const [pwa] = await Pwa.findOrCreate({ where: { publicationId, domain } })
  pwa.config = await import('./data/pwa/config.json') as PwaConfig
  pwa.design = await import('./data/pwa/design.json') as PwaDesign
  const layout = await import('./data/pwa/layout.json') as PwaLayout
  Object.assign(layout.columns[0].blocks[0].config, { issueIds: issues.map(({ id }) => id) })
  Object.assign(layout.columns[0].blocks[1].config, { title: categoryA.title, categoryId: categoryA.id })
  Object.assign(layout.columns[0].blocks[2].config, { title: categoryB.title, categoryId: categoryB.id })
  Object.assign(layout.columns[1].blocks[3].config, { title: categoryC.title, categoryId: categoryC.id })
  pwa.layout = layout
  await pwa.save()

  // Create PWA Optin
  console.info('~> Updating Optin')
  const [pwaOptin] = await PwaOptin.findOrCreate({ where: { pwaId: pwa.id } })
  const pwaOptinData = await import('./data/pwa-optin/data.json')
  pwaOptin.title = pwaOptinData.title
  pwaOptin.description = pwaOptinData.description
  pwaOptin.config = pwaOptinData.config
  await pwaOptin.save()

  // Assign Articles to Optin
  console.info('~> Assigning Articles to Optin')
  for (const article of articles.slice(0, 10)) {
    article.pwaOptinId = pwaOptin.id
    if (article.changed()) { await article.save() }
  }
  await close()
}

main().catch(console.error)
