import { Reader, User } from '@magloft/db-client'
import { Res } from '@tsed/common'
import { Configuration, Injectable } from '@tsed/di'
import { Unauthorized } from '@tsed/exceptions'
import * as bcrypt from 'bcrypt'
import { createCipheriv, createDecipheriv, createHash, createHmac, pbkdf2Sync, randomBytes } from 'crypto'
import * as md5 from 'md5'

export interface AuthConfiguration {
  token: { secret: string }
  cookie: { secret: string, name: string, domain: string }
  legacy: { secret: string, secretKeyBase: string }
}

export type AuthRole = 'admin' | 'user' | 'anonymous'

export interface LegacyCookieData {
  'session_id': string
  'warden.user.user.key': [[number], string]
  'warden.user.user.session': { 'last_request_at': number }
}

@Injectable()
export class AuthService {
  private config: AuthConfiguration
  private tokenSecret: Buffer
  private cookieSecret: Buffer
  private cookieSignedSecret: Buffer

  constructor(@Configuration() configuration: Configuration) {
    this.config = configuration.get<AuthConfiguration>('auth')
    this.tokenSecret = pbkdf2Sync(this.config.token.secret, 'magloft auth', 1000, 32, 'SHA1')
    this.cookieSecret = pbkdf2Sync(this.config.cookie.secret, 'encrypted cookie', 1000, 32, 'SHA1')
    this.cookieSignedSecret = pbkdf2Sync(this.config.cookie.secret, 'signed encrypted cookie', 1000, 64, 'SHA1')
  }

  validatePassword(encryptedPassword: string, password: string): boolean {
    return bcrypt.compareSync(password, encryptedPassword)
  }

  encryptPassword(password: string): string {
    const salt = bcrypt.genSaltSync(10, 'a')
    return bcrypt.hashSync(password, salt)
  }

  getDefaultPassword(token: string, date: Date): string {
    const timestamp = Math.floor(date.getTime() / 1000)
    const hash = md5(`${token}-${timestamp}-${this.config.legacy.secretKeyBase}`)
    return hash.substr(0, 7)
  }

  public async decryptLegacyHeaderToken(accessToken: string) {
    const [userIdString, digest] = accessToken.split('$')
    if (!userIdString || !digest) { return null }
    const userId = parseInt(userIdString, 10)
    const hash = createHash('sha1')
    hash.update(`${this.config.legacy.secret}-${userId}`)
    const targetDigest = hash.digest('hex').substr(7, 16)
    if (digest !== targetDigest) { return null }
    const user = await User.findById(userId)
    if (!user) { return null }
    return user
  }

  public encrypt(user: User | Reader) {
    const iv16 = Buffer.from(user.encryptedPassword.substring(0, 16))
    const cipher = createCipheriv('aes-256-cbc', this.tokenSecret, iv16)
    const hmac = createHmac('sha1', this.tokenSecret)
    const payload = Buffer.concat([cipher.update(String(user.id)), cipher.final()])
    hmac.update(payload)
    const digest = hmac.digest().slice(4, 8)
    const token = Buffer.concat([payload, iv16, digest])
    return token.toString('hex')
  }

  public async decrypt<T extends User | Reader>(tokenString: string, tokenType: 'reader' | 'user' = 'user', resolver?: (userId: number) => Promise<T>): Promise<T | null> {
    const token = Buffer.from(tokenString, 'hex')
    const digest = token.slice(-4)
    const iv16 = token.slice(-20, -4)
    const payload = token.slice(0, -20)
    const hmac = createHmac('sha1', this.tokenSecret)
    hmac.update(payload)
    const targetDigest = hmac.digest().slice(4, 8)
    if (Buffer.compare(digest, targetDigest) !== 0) { return null }
    const cipher = createDecipheriv('aes-256-cbc', this.tokenSecret, iv16)
    const result = Buffer.concat([cipher.update(payload), cipher.final()])
    const userId = parseInt(result.toString('utf8'), 10)
    const user: T = resolver
      ? await resolver(userId)
      : (tokenType === 'reader' ? await Reader.findById(userId) : await User.findById(userId)) as T
    if (!user) { return null }
    const targetIv16 = Buffer.from(user.encryptedPassword.substring(0, 16))
    if (Buffer.compare(iv16, targetIv16) !== 0) { return null }
    return user as unknown as T
  }

  public async decryptCookie(cookie: string) {
    const [messageEncoded, digest] = cookie.split('--')
    const hmac = createHmac('sha1', this.cookieSignedSecret)
    hmac.update(messageEncoded)
    if (digest !== hmac.digest('hex')) { throw new Unauthorized('Access denied') }
    const message = Buffer.from(messageEncoded, 'base64').toString('utf8')
    const [payload, iv] = message.split('--').map((part) => Buffer.from(part, 'base64'))
    const cipher = createDecipheriv('aes-256-cbc', this.cookieSecret, iv)
    const part = Buffer.from(cipher.update(payload)).toString('utf8')
    const final = cipher.final('utf8')
    const result = [part, final].join('')
    const authCookieData: LegacyCookieData = JSON.parse(result)
    if (!('warden.user.user.key' in authCookieData)) { return null }
    const userId = authCookieData['warden.user.user.key'][0][0]
    const encryptedPassword = authCookieData['warden.user.user.key'][1]
    const user = await User.findById(userId)
    if (!user) { return null }
    if (user.encryptedPassword.substring(0, 29) !== encryptedPassword) { return null }
    return user
  }

  public encryptCookie(user: User | Reader): string {
    const sessionId = randomBytes(16).toString('hex')
    const result = JSON.stringify({
      'session_id': sessionId,
      'warden.user.user.key': [[user.id], user.encryptedPassword.substring(0, 29)],
      'warden.user.user.session': { 'last_request_at': Math.floor(new Date().getTime() / 1000) }
    })
    const iv = randomBytes(16)
    const cipher = createCipheriv('aes-256-cbc', this.cookieSecret, iv)
    const payload = Buffer.concat([cipher.update(result), cipher.final()])
    const message = `${payload.toString('base64')}--${iv.toString('base64')}`
    const messageEncoded = Buffer.from(message, 'utf8').toString('base64')
    const hmac = createHmac('sha1', this.cookieSignedSecret)
    hmac.update(messageEncoded)
    const digest = hmac.digest('hex')
    return `${messageEncoded}--${digest}`
  }

  public generateFriendlyToken(length = 20) {
    const rlength = (length * 3) / 4
    const token = randomBytes(rlength).toString('base64')
    const base = { 'l': 's', 'I': 'x', 'O': 'y', '0': 'z', '/': '_', '+': '-', '=': '' }
    return token.replace(/[lIO0/+=]/g, m => base[m])
  }

  public setCookie(response: Res, user: User | Reader) {
    response.cookie(this.cookieName, this.encryptCookie(user), { domain: this.config.cookie.domain, sameSite: 'none', secure: true })
  }

  public clearCookie(response: Res) {
    response.clearCookie(this.cookieName, { domain: this.config.cookie.domain })
  }

  public get cookieName() {
    return this.config.cookie.name
  }
}
