import { Publication, Reader } from '@magloft/db-client'
import { EmailTemplate } from '@magloft/db-client/lib/models/EmailTemplate'
import { Configuration, Injectable } from '@tsed/di'
import * as nodemailer from 'nodemailer'
import { htmlToText } from 'nodemailer-html-to-text'
import * as mg from 'nodemailer-mailgun-transport'
import Mail from 'nodemailer/lib/mailer'
import { EmailTemplateLinks, EmailTemplateVariables, parseEmailTemplate } from '../lib/email'
import { AuthService } from './AuthService'

export interface MailgunConfiguration {
  auth: {
    api_key: string
    domain: string
  }
}

@Injectable()
export class MailService {
  private config: MailgunConfiguration
  private transport: Mail

  constructor(private auth: AuthService, @Configuration() configuration: Configuration) {
    this.config = configuration.get<MailgunConfiguration>('mailgun')
    this.transport = nodemailer.createTransport(mg(this.config))
    this.transport.use('compile', htmlToText())
  }

  sendReaderEmail(publication: Publication, reader: Reader, emailTemplate: EmailTemplate, extraVariables: EmailTemplateVariables = {}) {

    const variables: EmailTemplateVariables = {
      ...emailTemplate.variables,
      'reader.name': reader.name,
      'reader.lastname': reader.lastname,
      'reader.email': reader.email,
      'reader.password': reader.hasDefaultPassword ? this.auth.getDefaultPassword(reader.email, reader.createdAt) : '',
      'banner': `https://cdn.magloft.com/${publication.banners[0]['image']}`,
      'appicon': `https://cdn.magloft.com/${publication.properties.get('appIcon')}`,
      ...extraVariables
    }
    const links: EmailTemplateLinks = {
      'confirmation_url': `https://www.magloft.com/readers/confirmation?confirmation_token=${reader.confirmationToken}`,
      'edit_password_url': `https://www.magloft.com/readers/password/edit?reset_password_token=${reader.resetPasswordToken}`,
      'launch_app_url': `https://www.magloft.com/launch/${publication.appId}`
    }
    return this.sendMail({
      from: `"${publication.title}" <support@magloft.com>`,
      to: reader.email,
      subject: emailTemplate.subject,
      html: parseEmailTemplate(emailTemplate.html, variables, links)
    })
  }

  sendMail(options: Mail.Options) {
    return this.transport.sendMail(options)
  }
}
