import * as db from '@magloft/db-client'
import { Issue, Pwa, Reader, ReaderPurchase, ReaderSubscription } from '@magloft/db-client'
import { Configuration, Injectable } from '@tsed/di'
import { UnprocessableEntity } from '@tsed/exceptions'
import { CheckoutSession, PaymentProvider, PaymentProviderStatic, PaymentScope, PriceInterval, Product, Unlock, UnlockKind } from '../lib/payments/PaymentProvider'
import { StripePaymentProvider } from '../lib/payments/StripePaymentProvider'

export interface PurchaseIssueParams {
  source: string
  issueId: number
  publicationId: number
  readerId: number
  receipt?: object
  orderId?: string
}

export interface PurchaseSubscriptionParams {
  source: string
  orderId: string
  appSubscriptionId: number
  publicationId: number
  readerId: number
  receipt?: object
}

export class PurchaseExistsError extends UnprocessableEntity {
  name = 'PURCHASE_EXISTS'
  constructor(public kind: string, public id?: string) {
    super(`${kind} already exists`)
  }
}

@Injectable()
export class PaymentService {
  private providers = new Map<string, PaymentProviderStatic>([
    ['stripe', StripePaymentProvider]
  ])

  constructor(@Configuration() private configuration: Configuration) { }

  getProvider(pwa: Pwa, identifier = 'stripe') {
    const Provider = this.providers.get(identifier)
    if (!Provider) { throw new Error(`Invalid payment provider: ${identifier}`) }
    return new Provider(this.configuration, pwa)
  }

  async syncScope(provider: PaymentProvider, { kind, publication, id }: PaymentScope): Promise<Product> {
    if (kind === 'issue') {
      // Retrieve Issue
      const issue = await db.Issue.findOne({ where: { publicationId: publication.id, id, unlockType: 'paid', date: { [db.Op.ne]: null } } })
      if (!issue) { throw new Error('Issue not found') }

      // Retrieve Product
      return provider.syncProduct({
        id: `pwa-issue-${issue.id}`,
        name: issue.title,
        description: issue.info,
        images: [`https://cdn.magloft.com/${issue.cover}`],
        price: { currency: 'usd', interval: PriceInterval.ONCE, amount: issue.priceTier - 0.01 },
        unlocks: [{ kind: UnlockKind.ISSUE, id }]
      })
    } else if (kind === 'subscription') {
      // Retrieve Subscription
      const subscription = await db.AppSubscription.findOne({ where: { publicationId: publication.id, id } })
      if (!subscription) { throw new Error('Subscription not found') }

      // Retrieve Product
      const duration = subscription.duration.split('.')
      const interval = duration[0] as PriceInterval
      const intervalCount = parseInt(duration[1], 10)
      return provider.syncProduct({
        id: `pwa-subscription-${subscription.id}`,
        name: subscription.title,
        description: subscription.description,
        images: [`https://cdn.magloft.com/${publication.properties.get('appIcon')}`],
        price: { currency: 'usd', interval, intervalCount, amount: subscription.price - 0.01 },
        unlocks: [{ kind: UnlockKind.SUBSCRIPTION, id: subscription.id }]
      })
    }
  }

  completeCheckoutSession(session: CheckoutSession) {
    return Promise.all(session.product.unlocks.map((unlock) => {
      return this.completeCheckoutSessionUnlock(session, unlock)
    }))
  }

  async completeCheckoutSessionUnlock({ publicationId, readerId, extra, orderId }: CheckoutSession, { kind, id }: Unlock) {
    if (kind === 'issue') {
      await this.purchaseIssue({ issueId: id, publicationId, readerId, source: 'pwa', orderId })
    } else if (kind === 'subscription') {
      await this.purchaseSubscription({ appSubscriptionId: id, publicationId, readerId, source: 'pwa', orderId })
      if (extra) { await this.purchaseIssue({ issueId: extra, publicationId, readerId, source: 'pwa', orderId }) }
    }
  }

  async ensurePurchaseAvailable(reader: Reader, { publication, kind, id }: PaymentScope): Promise<void> {
    const hasSubscription = await this.unlockSubscriptionPurchases(reader, publication.id)
    if (kind === 'subscription' && hasSubscription) { throw new PurchaseExistsError('subscription') }
    if (kind === 'issue') {
      const purchaseCount = await ReaderPurchase.count({ where: { publicationId: publication.id, readerId: reader.id, issueId: id } })
      if (purchaseCount > 0) { throw new PurchaseExistsError('issue') }
    }
  }

  async unlockSubscriptionPurchases(reader: Reader, publicationId: number): Promise<boolean> {
    const subscriptions = await reader.getReaderSubscriptions({ where: { active: true } })
    if (subscriptions.length === 0) { return false }
    const [start] = subscriptions.filter(({ startDate }) => startDate).map(({ startDate }) => new Date(startDate).setHours(0, 0, 0, 0)).sort((a, b) => a - b)
    const [end] = subscriptions.filter(({ endDate }) => endDate).map(({ endDate }) => new Date(endDate).setHours(23, 59, 59, 0)).sort((a, b) => b - a)
    const paidWhere: db.WhereOptions = { unlockType: 'paid' }
    if (start && end) {
      paidWhere.date = { [db.Op.between]: [start, end] }
    } else if (start) {
      paidWhere.date = { [db.Op.gte]: start }
    } else if (end) {
      paidWhere.date = { [db.Op.lte]: end }
    }
    const issues = await Issue.findAll({ where: { publicationId, [db.Op.or]: [{ unlockType: 'subscribe' }, paidWhere] } })
    if (issues.length > 0) { return true }
    const latestIssue = await Issue.findOne({ where: { publicationId, unlockType: 'paid' }, order: [['date', 'desc']], limit: 1 })
    if (!latestIssue) { return true }
    await this.purchaseIssue({ publicationId, readerId: reader.id, issueId: latestIssue.id, source: 'subscription' })
    return true
  }

  purchaseIssue({ publicationId, readerId, issueId, source, receipt, orderId }: PurchaseIssueParams) {
    return ReaderPurchase.findOrCreate({ where: { publicationId, readerId, issueId }, defaults: { receipt, orderId, verified: true, source } })
  }

  purchaseSubscription({ publicationId, readerId, appSubscriptionId, source, receipt, orderId }: PurchaseSubscriptionParams) {
    return ReaderSubscription.findOrCreate({ where: { publicationId, readerId, appSubscriptionId }, defaults: { receipt, orderId, verified: true, active: true, startDate: new Date(), source } })
  }
}
