import { PushDataChrome, PushDataSafari, PushProvider, Pwa, PwaPushSubscription } from '@magloft/db-client'
import { Injectable } from '@tsed/di'
import * as md5 from 'md5'
import { RequestOptions, sendNotification } from 'web-push'
import { CreatePushDeliveryParams } from '../api/pwa/PushController'

export interface PushConfig {
  google: {
    serverPublicKey: string
    serverPrivateKey: string
  }
  apple: {
    key: string
    certificate: string
  }
}

@Injectable()
export class PushService {
  getDomainFromWebsitePushId(websitePushId: string) {
    const [, ...parts] = websitePushId.split('.')
    return parts.reverse().join('.')
  }

  sendNotification(params: CreatePushDeliveryParams, subscription: PwaPushSubscription, pwa: Pwa) {
    if (subscription.provider === 'chrome') {
      return this.sendNotificationChrome(params, subscription, pwa)
    } else if (subscription.provider === 'safari') {
      return this.sendNotificationSafari(params, subscription, pwa)
    } else {
      throw new Error('Unable to send notification: invalid provider')
    }
  }

  getToken(provider: PushProvider, data: PushDataChrome | PushDataSafari) {
    if (provider === 'chrome') { return md5((data as PushDataChrome).keys.auth) }
    if (provider === 'safari') { return md5((data as PushDataSafari).deviceToken) }
    throw new Error(`Invalid push provider: ${provider}`)
  }

  private sendNotificationChrome({ title, body, data, actions }: CreatePushDeliveryParams, subscription: PwaPushSubscription<PushDataChrome>, pwa: Pwa) {
    const { google: { serverPrivateKey, serverPublicKey } } = pwa.config.push as PushConfig
    const options: RequestOptions = { vapidDetails: { subject: 'mailto:office@magloft.com', publicKey: serverPublicKey, privateKey: serverPrivateKey } }
    const payload = JSON.stringify({ notification: { title, body, data, actions, icon: `https://${pwa.domain}/assets/icons/icon-512x512.png` } })
    return sendNotification(subscription.data, payload, options).then(() => true).catch(() => false)
  }

  private sendNotificationSafari({ title, body, data, actions }: CreatePushDeliveryParams, subscription: PwaPushSubscription<PushDataSafari>, pwa: Pwa) {
    return Promise.reject(new Error('Not implemented'))
  }

  encodeAuthenticationToken(pwaId: number, readerId?: number) {
    const buffer = new Buffer(8)
    buffer.writeUInt32BE(pwaId, 0)
    buffer.writeUInt32BE(readerId ?? 0, 4)
    return buffer.toString('hex')
  }

  decodeAuthenticationToken(token: string) {
    const buffer = Buffer.from(token, 'hex')
    const pwaId = buffer.readUInt32BE(0)
    const readerId = buffer.readUInt32BE(4)
    return { pwaId, readerId: readerId === 0 ? null : readerId }
  }
}
