
import { $log } from '@tsed/common'
import { Configuration, Injectable } from '@tsed/di'
import Stripe from 'stripe'

export interface InvoiceData {
  customer: Stripe.Customer
  amount: number
  currency: 'usd'
  description: string
  collectionMethod: 'send_invoice' | 'charge_automatically'
  daysUntilDue?: number
}

export interface StripeConfiguration {
  secretKey: string
  webhookSecrets: { [key: string]: string }
}

@Injectable()
export class StripeService {
  private stripe: Stripe
  private config: StripeConfiguration

  constructor(@Configuration() private configuration: Configuration) {
    this.config = this.configuration.get<StripeConfiguration>('stripe')
    this.stripe = new Stripe(this.config.secretKey, { apiVersion: '2020-08-27' })
  }

  async getCustomer(id: string) {
    return this.stripe.customers.retrieve(id) as unknown as Stripe.Customer
  }

  async createCustomer(params: Stripe.CustomerCreateParams) {
    return this.stripe.customers.create(params)
  }

  async createInvoice(params: InvoiceData) {
    // Create Invoice Item
    const invoiceItem = await this.stripe.invoiceItems.create({
      customer: params.customer.id,
      amount: params.amount,
      currency: params.currency,
      description: params.description
    })

    // Create Invoice (this will pull all pending invoice items)
    const invoice = await this.stripe.invoices.create({
      'days_until_due': params.daysUntilDue,
      'collection_method': params.collectionMethod,
      'customer': invoiceItem.customer as string,
      'metadata': { type: 'flexpdf-credit' }
    })

    // Finalize Invoice
    return this.stripe.invoices.finalizeInvoice(invoice.id)
  }

  async payInvoice(invoice: Stripe.Invoice) {
    if (invoice.collection_method === 'charge_automatically') {
      return this.stripe.invoices.pay(invoice.id)
    } else {
      return this.stripe.invoices.sendInvoice(invoice.id)
    }
  }

  validateEvent(scope: string, signature: string, payload: string | Buffer) {
    try {
      const webhookSecret = this.config.webhookSecrets[scope]
      return this.stripe.webhooks.constructEvent(payload, signature, webhookSecret)
    } catch (error) {
      $log.error(error)
      return null
    }
  }
}
