import { Reader, User } from '@magloft/db-client'
import { AuthService } from '../../src/services/AuthService'
import { bootstrap, createInjectModule, mockResponse } from '../helpers'

interface Context { service: AuthService }

const AuthServiceModule = createInjectModule<Context, typeof AuthService>('service', AuthService)

describe('AuthService', () => {
  const ctx = bootstrap<Context>([AuthServiceModule])

  describe('auth', () => {
    it('encrypts and decrypts a password', () => {
      const encryptedPassword = ctx.service.encryptPassword('test')
      const success = ctx.service.validatePassword(encryptedPassword, 'test')
      expect(success).toEqual(true)
    })

    it('retrieves a default password', () => {
      const date = new Date(0)
      const password = ctx.service.getDefaultPassword('test@magloft.com', date)
      expect(password).toEqual('60db8f7')
    })

    it('encrypts and decrypts a user via token', async () => {
      const userA = await User.findById(1)
      const token = ctx.service.encrypt(userA)
      const userB = await ctx.service.decrypt<User>(token, 'user')
      expect(userA.id).toEqual(userB.id)
      const userC = await ctx.service.decrypt<User>('a7400db8d1999ae13257e6d0806ecf36243261243130245a437664484d33782ee0af7eb0', 'user')
      expect(userC).toBeNull()
    })

    it('encrypts and decrypts a reader via token', async () => {
      const readerA = await Reader.findById(1)
      const token = ctx.service.encrypt(readerA)
      const readerB = await ctx.service.decrypt<Reader>(token, 'reader')
      expect(readerA.id).toEqual(readerB.id)
      const readerC = await ctx.service.decrypt<Reader>('a7400db8d1999ae13257e6d0806ecf36243261243130245a437664484d33782ee0af7eb0', 'reader')
      expect(readerC).toBeNull()
      const readerD = await ctx.service.decrypt<Reader>('82492bca0411edf32d59da5bc714a810243261243130246a554579454e4677546f84326b', 'reader')
      expect(readerD).toBeNull()
      const readerE = await ctx.service.decrypt<Reader>('e32bc94ce9e3e1b6a859115c18cee6cd2432612431302461663276344f51676223cf0140', 'reader')
      expect(readerE).toBeNull()
    })

    it('decrypts legacy tokens', async () => {
      const userA = await ctx.service.decryptLegacyHeaderToken('1$8e6efcf8e6d98513')
      expect(userA.id).toEqual(1)
      const userB = await ctx.service.decryptLegacyHeaderToken('1$')
      expect(userB).toBeNull()
      const userC = await ctx.service.decryptLegacyHeaderToken('1$abc')
      expect(userC).toBeNull()
      const userD = await ctx.service.decryptLegacyHeaderToken('19243$ac9a8de2251db5a6')
      expect(userD).toBeNull()
    })

    it('encrypts and decrypts a user via cookies', async () => {
      const userA = await User.findById(1)
      const cookie = ctx.service.encryptCookie(userA)
      const userB = await ctx.service.decryptCookie(cookie)
      expect(userA.id).toEqual(userB.id)
    })

    it('generates a friendly token', () => {
      const token = ctx.service.generateFriendlyToken(16)
      expect(token).toHaveLength(16)
    })

    it('sets a cookie', async () => {
      const response = mockResponse()
      const user = await User.findById(1)
      ctx.service.setCookie(response, user)
      expect(response.cookie).toHaveBeenCalled()
    })

    it('clears a cookie', () => {
      const response = mockResponse()
      ctx.service.clearCookie(response)
      expect(response.clearCookie).toHaveBeenCalled()
    })
  })
})
